use strict;

package SoMat::DB::Tab;

use Symbol;

sub new {
    my ($class, $fname) = @_;
    my $s = {};
    bless $s, ref $class || $class;
    $s->{separator} = "\t";
    $s->{rows} = [];
    $s->{currentrow} = 0;
    $s->{headidx} = {};
    $s->{heads} = [];
    $s->{filename} = $fname if defined $fname;
    return $s;
}

sub read_line {
    my ($s, $fh) = @_;
    my $line = <$fh>;
    return undef unless defined $line;
    chomp $line;
    return [ split($s->{separator}, $line, -1) ];
}

sub read { # KLUDGE doesn't do locking
    my ($s, $fname) = @_;
    $fname = $s->{filename} unless defined $fname;
    return undef unless defined $fname;
    return $s->_read($fname) if ref $fname;
    my $fh = gensym;
    open($fh, "<$fname") or return undef;
    $s->_read($fh);
    close($fh); # KLUDGE check errors
    return 1;
}

sub _read {
    my ($s, $fh) = @_;
    while (defined (my $row = $s->read_line($fh))) {
        push(@{$s->{rows}}, $row);
    }
    $s->set_heads($s->get_row(0));
    return 1;
}

sub write { # KLUDGE doesn't do locking
    my ($s, $fname) = @_;
    $fname = $s->{filename} unless defined $fname;
    return undef unless defined $fname;
    return $s->_write($fname) if ref $fname;
    my $fh = gensym;
    open($fh, ">$fname") or return undef;
    $s->_write($fh);
    close($fh); # KLUDGE check errors
    return 1;
}

sub _write {
    my ($s, $fh) = @_;
    my $sep = $s->{separator};
    for my $row (@{$s->{rows}}) {
        print $fh join($sep, @$row), "\n";
    }
    return 1;
}

sub get_row_count {
    my ($s) = @_;
    return scalar @{$s->{rows}};
}

sub get_row {
    my ($s, $row) = @_;
    $row = $s->{currentrow} unless defined $row;
    return $s->{rows}->[$row];
}

sub seek_row {
    my ($s, $row) = @_;
    return undef if $row < 0 || $row >= $s->get_row_count;
    return $s->{currentrow} = $row;
}

sub seek_field {
    my ($s, $field, $value) = @_;
    if (ref($field) eq 'HASH') {
        ROWS: for my $i (1 .. @{$s->{rows}} - 1) {
            $s->seek_row($i);
            while (my ($key, $val) = each %$field) {
                next ROWS if $val ne $s->get_field($key);
            }
            return $i;
        }
        return undef;
    }
    for my $i (1 .. @{$s->{rows}} - 1) {
        $s->seek_row($i);
        return $i if $value eq $s->get_field($field);
    }
    return undef;
}

sub get_col {
    my $s = shift;
    my @cols;
    my $row = $s->get_row;
    for my $col (@_) {
        push(@cols, $row->[$col]);
    }
    return \@cols;
}

sub get_field {
    my $s = shift;
    my $row = $s->get_row;
    my @cols;
    for my $field (@_) {
        push(@cols, (exists $s->{headidx}->{$field})
            ? $row->[$s->{headidx}->{$field}] : undef);
    }
    return (@cols > 1) ? \@cols : $cols[0];
}

sub set_field {
    my $s = shift;
    my $row = $s->get_row;
    while (@_) {
        my $field = shift;
        my $value = shift;
        $s->add_head($field); # KLUDGE BAD
        $row->[$s->{headidx}->{$field}] = $value;
    }
    return 1;
}

sub get_row_hash {
    my ($s, $rn) = @_;
    my $row = $s->get_row($rn);
    my %r;
    my $i = 0;
    for my $h (@{$s->{heads}}) {
        $r{$h} = $row->[$i++];
    }
    return \%r;
}

sub add_head {
    my $s = shift;
    for my $head (@_) {
        next if exists $s->{headidx}->{$head};
        my $i = scalar @{$s->{heads}};
        $s->{heads}->[$i] = $head;
        $s->{headidx}->{$head} = $i;
    }
    return 1;
}

sub set_heads {
    my $s = shift;
    $s->{heads} = $s->{rows}->[0] = shift;
    my $i = 0;
    foreach my $head (@{$s->{heads}}) {
        $s->{headidx}->{$head} = $i++;
    }
}

sub splice_rows {
    my $s = shift;
    return splice(@{$s->{rows}}, @_);
}

sub append_row {
    my ($s, $row) = @_;
    $row = [] unless defined $row;
    push(@{$s->{rows}}, $row);
    $s->seek_row($s->get_row_count - 1);
    return $row;
}

1;
