package SoMat::Network::Mail;

use strict;
use Symbol ();
use Socket;

sub new {
    my ($proto, $args) = @_; # args: mailserver from to cc bcc subject
    my $s = bless Symbol::gensym(), ref($proto) || $proto;
    $s->set('timeout', 15);
    for my $k (keys %$args) {
        $s->set($k, $args->{$k});
    }
    $s->connect($s->get('mailserver'), 25);
    $s->start_message();
    return $s;
}

sub debug {
    #return;
    my $s = shift;
    my $fh = Symbol::gensym();
    open($fh, ">>/tmp/mail.log") or die "Can't open debug file: $!";
    print $fh @_;
    close($fh);
}

sub set {
    my ($s, $name, $value) = @_;
    ${*$s}{$name} = $value;
}

sub get {
    my ($s, $name) = @_;
    return ${*$s}{$name};
}

sub DESTROY {
    my ($s) = @_;
    $s->end_message();
    $s->close();
}

sub start_message {
    my ($s) = @_;
$s->debug("start_message IN, mailserver = ", $s->get('mailserver'), "\n");
    $s->check_response(220);
    $s->sendf("HELO %s", $s->get('mailserver'));
    $s->check_response(250);
    $s->sendf("MAIL From: %s", _fix_name($s->get('from')));
    $s->check_response(250);
    my $rcpts = $s->_join_names($s->get('to'), $s->get('cc'), $s->get('bcc'));
    for my $recipient (@$rcpts) {
        $s->sendf("RCPT To: %s", _fix_name($recipient));
        $s->check_response(250);
    }
    $s->sendf("DATA");
    $s->check_response(354);
$s->debug("start_message OUT\n");
}

sub send_headers {
    my ($s) = @_;
    $s->sendf("From: %s", $s->_expand_names($s->get('from')));
    $s->sendf("To: %s", $s->_expand_names($s->get('to')));
    if (my $cc = $s->get('cc')) {
        $s->sendf("Cc: %s", $s->_expand_names($cc));
    }
    if (my $subject = $s->get('subject')) {
        $s->sendf("Subject: $subject");
    }
    $s->sendf("");
}

sub end_message {
    my ($s) = @_;
    $s->sendf("");
    $s->sendf(".");
    $s->check_response(250);
    $s->sendf("QUIT");
    $s->check_response(221);
}

sub connect {
    my ($s, $remote, $port) = @_;
$s->debug("connect IN\n");
    $remote = 'localhost' unless defined $remote;
    $port = 25 unless defined $port;
    $port = getservbyname($port, 'tcp') if $port =~ /\D/;

$s->debug("connect 1\n");
    die "No port" unless $port;
    my $iaddr = inet_aton($remote) or die "no host: $remote";
    my $paddr = sockaddr_in($port, $iaddr);
    my $proto = getprotobyname('tcp');
$s->debug("connect 2\n");
    socket($s, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
    local $SIG{ALRM} = sub { die "Mail: connect() to $remote timed out\n" };
    alarm $s->get('timeout');
    CORE::connect($s, $paddr) or die "connect: $!";
    alarm 0;
    my $fh = select $s;
    $| = 1;
    select $fh;
$s->debug("connect OUT\n");
}

sub close {
    my ($s) = @_;
    CORE::close($s);
}

sub print {
    my $s = shift;
$s->debug("SENT: ", @_);
    CORE::print $s @_;
}

sub printf {
    my $s = shift;
$s->debug("SENT: ", sprintf(@_));
    CORE::printf $s @_;
}

sub sendf {
    my $s = shift;
    my $fmt = shift;
$s->debug("SENT: ", sprintf($fmt . "\015\012", @_));
    CORE::printf $s $fmt . "\015\012", @_;
}

sub check_response {
    my ($s, $code) = @_;
$s->debug("check_response IN, code = $code\n");
    while (1) {
        my $line = <$s>;
$s->debug("RESPONSE: $line");
        if ($line =~ /^(\d\d\d)(.?)/) {
            if ($1 != $code) {
                die "SMTP Error - Expected code $code, got: $line";
            }
            next if $2 eq '-';
            return;
        }
        die "SMTP Error - Bad response: $line";
    }
}

sub _fix_name {
    return map { /(\<.*\>)/ } @_;
}

sub _expand_names {
    my ($s, $names) = @_;
    if (ref $names eq 'ARRAY') {
        return join(", ", @$names);
    }
    if (!ref $names) {
        return $names;
    }
    return undef;
}

sub _join_names {
    my $s = shift;
    my @ret;
    for my $names (@_) {
        if (ref $names eq 'ARRAY') {
            for my $subname (@$names) {
                push(@ret, $subname);
            }
        }
        if (!ref $names && $names) {
            push(@ret, $names);
        }
    }
    return \@ret;
}

1;
