package SoMat::Network::ProcessMonitor;

use Fcntl;
require SoMat::Network::Dispatch;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    $s->{inputs} = {};
    $s->{name_order} = [];
    $s->{old_message} = {};
    $s->{counts} = {};
    $s->{update_interval} ||= 1;
    $s->{dispatcher} = SoMat::Network::Dispatch->new;
    return $s;
}

sub fork_updater {
    my ($s, $code, $tag) = @_;
    my ($readfh, $writefh);
    pipe($readfh, $writefh);
    my $pid = fork();
    die "fork() failed: $!" unless defined $pid;
    if ($pid) {
        close($writefh);
        my $flags = fcntl($readfh, F_GETFL, 0);
        $flags = fcntl($readfh, F_SETFL, $flags | O_NONBLOCK);
        $s->{dispatcher}->add('R', $readfh, sub {
            my ($fh, $data) = @_;
            my $buf = '';
            while (1) {
                my $got = sysread($fh, $buf, 2048, length($buf));
                unless ($got) {
                    $s->{dispatcher}->remove($fh) if defined $got;
                    last;
                }
                #print "GOT ($data->{name}): $buf\n";
                $data->{buf} .= $buf;
                while ($data->{buf} =~ s/^(.*?)\n//) {
                    $s->monitor($data, $1);
                }
            }
        }, $tag);
    } else {
        close($readfh);
        my $fd = fileno($writefh);
        open(STDOUT, ">&=$fd") or die "Can't open STDOUT: $!";
        $| = 1;
        eval { &$code(); };
        my $type = ($@) ? 'fatal' : 'finished';
        my $err = $@ || 'Update complete';
        my $msg = {type=>$type, message=>$err};
        for my $k (keys %$tag) {
            $msg->{$k} ||= $tag->{$k};
        }

        # KLUDGE copied from SoMat::eDAQ::Hardware::Update::emit_message()
        my @out;
        while (my ($k, $v) = each %$msg) {
            $v =~ s/([\\"\x00-\x19])/sprintf('\\x%02X', ord($1))/seg;
            push @out, qq{"$k"=>"$v"};
        }
        print "{", join(",", @out), "}\n";

        while (wait() != -1) {
            # Get rid of zombies so we can really exit...
        }
        exit(0);
    }
}

sub fatal_count {
    my ($s) = @_;
    return $s->{counts}{fatal};
}

sub monitor {
    my ($s, $in, $line) = @_;
    my $old_in = $s->{inputs}{$in->{name}};
    unless ($old_in) {
        $old_in = {};
        push @{$s->{name_order}}, $in->{name};
    }
    $line = '' unless defined $line;
    $line =~ s/\s+$//s;
    $in->{last_line} = $line;
    if ($line =~ /^\{/) {
        if (my $prog = eval $line) {
            $prog->{type} ||= 'progress';
            $prog->{progress} ||= $old_in->{progress};
            map { $in->{$_} = $prog->{$_} if exists $prog->{$_} }
                qw(type done todo time message progress);
        }
    } elsif ($line ne '') {
        $in->{debug} = $line;
    } else {
        # Just update with $in as-is
    }
    ++$s->{counts}{$in->{type} || 'undef'};
    $s->{inputs}{$in->{name}} = $in;

    my $message_changed =
        $in->{message} ne ($s->{old_message}{$in->{name}} || '');
    $s->{old_message}{$in->{name}} = $in->{message};

    my $elapsed = time - $s->{base_time};
    my $elapsed_changed = ($elapsed - $s->{elapsed}) >= $s->{update_interval};
    $s->{elapsed} = $elapsed if $elapsed_changed;

    if ($s->{comet}) {
        if ($in->{debug}) {
            print qq{<pre>$in->{debug}</pre>\n};
            $in->{debug} = undef;
            return;
        }

        my $update = "window.parent.comet_update(" . $s->to_json($in) . ");";
        print qq{<script type="text/javascript">\n$update\n</script>\n};

        if (0) {
            print qq{<pre>$update\n</pre>\n};
        }
    } else {
        if ($in->{debug}) {
            print "$in->{debug}\n";
            $in->{debug} = undef;
            return;
        }
        if ($message_changed && $in->{message}) {
            my $msg = $in->{message};
            if ($in->{type} eq 'fatal') {
                $msg = qq{<div class="errorbox">FATAL ERROR: $msg<br>}
                  . '<br>Please contact <a href="mailto:support@somat.com">'
                  . 'SoMat Customer Support</a> with the above message and the'
                  . ' contents of the <a href="/eDAQ/Test/Logbook">eDAQ LogBook'
                  . '</a> for assistance.'
                  . '</div>';

            } else {
                $msg =~ s/^([^\n]+)$/<tt>$1<\/tt>/s;
            }
            print "<tt>";
            printf "%d:%.02d -", int($elapsed / 60), ($elapsed % 60);
            printf " %s: </tt>%s<br>\n", $in->{name}, $msg;
        }
        return unless $elapsed_changed;
        my $out = sprintf "%d:%.02d - ", int($elapsed / 60), ($elapsed % 60);
        for my $name (@{$s->{name_order}}) {
            my $ti = $s->{inputs}{$name};
            $out .= sprintf("%s %3.0f%%    ", $ti->{name}, $ti->{progress});
        }
        $out =~ s/\s+$//;
        $out =~ s/ /&nbsp;/g;
        print "<tt>$out</tt><br>\n";
    }
}

sub to_json {
    my ($s, $o, $json) = @_;
    $json = '' unless defined $json;
    my $ref = ref($o);
    if (!$ref) {
        $o =~ s/(['\\])/\\$1/sg;
        $o =~ s/([\x00-\x19])/sprintf("\\u%04X", ord($1))/seg;
        $json .= "'$o'";
    } elsif ($ref eq 'ARRAY') {
        $json .= "[";
        my $comma = 0;
        for my $e (@$o) {
            $json .= ($comma++ ? ",\n" : "\n") . $s->to_json($e);
        }
        $json .= "\n]";
    } elsif ($ref eq 'HASH') {
        $json .= "{\n";
        my $comma = 0;
        for my $k (sort keys %$o) {
            $json .= ($comma++ ? ",\n" : "") . $s->to_json($k)
                . ": " . $s->to_json($o->{$k});
        }
        $json .= "\n}";
    }
    return $json;
}

1;
