package SoMat::Network::WebRequest;

use strict;
use Symbol ();
use Socket;

sub new {
    my $s = _new(@_);
    $s->parse_url($s->{url});
    $s->connect($s->{host}, $s->{port});
    if ($s->{post_form_data}) {
        $s->{method} = 'POST';
        $s->{body} = $s->escape_form_data($s->{post_form_data});
    }
    $s->do_request();
    $s->get_headers();
    return $s;
}

sub _new {
    my ($class, @args) = @_;
    unshift(@args, 'url') if @args == 1;
    my $s = bless { @args }, ref($class) || $class;
    $s->{method} ||= 'GET';
    $s->{socket} = Symbol::gensym();
    $s->{linebuf} = '';
    return $s;
}

sub DESTROY {
    my ($s) = @_;
#print STDERR "closing socket\n";
    $s->close();
}

sub get {
    my ($s, $name) = @_;
    return undef unless exists $s->{$name};
    return $s->{$name};
}

sub get_socket {
    return $_[0]->{socket};
}

sub parse_url {
    my ($s, $url) = @_;
    if ($url =~ m,http://([^/]+)(.*),) {
        ($s->{host}, $s->{path}) = ($1, $2);
        $s->{port} = 80;
        if ($s->{host} =~ m,([^:]+):(\d+),) {
            ($s->{host}, $s->{port}) = ($1, $2);
        }
#print STDERR "HOST: $s->{host}\n";
#print STDERR "PATH: $s->{path}\n";
#print STDERR "PORT: $s->{port}\n";
    } else {
        die "error parsing URL $url";
    }
}

sub escape_form_data {
    my ($s, $form_data) = @_;
    my @fd = (ref($form_data) eq 'HASH') ? %$form_data : @$form_data;
    my @efd;
    while (my ($k, $v) = splice(@fd, 0, 2)) {
        $v =~ s/([^A-Za-z0-9\-_.!~*'()])/sprintf("%%%02X", ord($1))/eg;
        push @efd, "$k=$v";
    }
    return join('&', @efd);
}

sub do_request {
    my ($s, $extra_headers) = @_;
    $extra_headers = '' unless defined $extra_headers;
    if (defined $s->{body}) {
        $extra_headers .= 'Content-length: ' . length($s->{body}) . "\r\n";
    }
    $s->print(
        "$s->{method} $s->{path} HTTP/1.0\r\n" .
        "Host: $s->{host}\r\n" .
        "User-Agent: WebRequest/0.0\r\n" .
        "$extra_headers\r\n"
    );
    $s->print($s->{body}) if defined $s->{body};
}

sub get_headers {
    my ($s) = @_;
#print STDERR "Getting headers\n";
    my $response = $s->getline();
    if ($response =~ m,HTTP/\S+ (\d+) (.*)\r\n,) {
        $s->{response_code} = $1;
        $s->{response_message} = $2;
#print STDERR "HEADERS CODE $1 MESSAGE $2\n";
    } else {
        die "ERROR: malformed response '$response': $!"
    }
    while (my $line = $s->getline()) {
        if ($line eq "\r\n") {
            return;
        } else {
            $s->{headers} .= $line;
        }
    }
}

sub get_response_code {
    return $_[0]->{response_code};
}

sub get_response_message {
    return $_[0]->{response_message};
}

sub connect {
    my ($s, $remote, $port) = @_;
    $remote = 'localhost' unless defined $remote;
    $port = 2345 unless defined $port; # edaqserver HTTP port
    if ($port =~ /\D/) { $port = getservbyname($port, 'tcp') }
    die "No port" unless $port;
    my $iaddr = inet_aton($remote) or die "no host: $remote";
    my $paddr = sockaddr_in($port, $iaddr);
    my $proto = getprotobyname('tcp');
    socket($s->{socket}, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
    connect($s->{socket}, $paddr) or die "connect: $!";
    my $fh = select $s->{socket};
    $| = 1;
    select $fh;
}

sub close {
    my ($s) = @_;
    CORE::close($s->{socket});
}

sub print {
    my $s = shift;
#print STDERR "SENT: ";
#print STDERR @_;
    print {$s->{socket}} @_;
}

##### New stuff

sub new_nonblock {
    my $s = _new(@_);
    return $s;
}

sub begin_connection {
    my ($s, $url) = @_;
    $s->{url} = $url if defined $url;
    $s->parse_url($s->{url});
    $s->connect($s->{host}, $s->{port});
    $s->do_request();
}

sub process_headers {
    my ($s) = @_;
    my $line;
    $s->{head_state} = 0 unless exists $s->{head_state};
    while (defined ($line = $s->getlinebuf)) {
        if ($s->{head_state} == 0) {
            if ($line =~ m,HTTP/\S+ (\d+) (.*)\r\n,) {
                $s->{response_code} = $1;
                $s->{response_message} = $2;
            } else {
                die "ERROR: malformed response '$line': $!"
            }
            ++$s->{head_state};
        } else {
            if ($line eq "\r\n") {
                return 1;
            } else {
                $s->{headers} .= $line;
            }
        }
    }
    return 0;
}

sub read {
    my ($s, $len) = @_;
    $len = 256 unless defined $len;
    my $got = sysread($s->{socket}, $s->{linebuf}, $len,
        length($s->{linebuf}));
    return $got;
}

sub getlinebuf {
    my ($s) = @_;
    if ($s->{linebuf} =~ /^([^\n]*\n)(.*)$/s) {
        $s->{linebuf} = $2;
        return $1;
    }
    return undef;
}

sub getline {
    my ($s) = @_;
    my $line = $s->getlinebuf;
    return $line if defined $line;
    my $got = $s->read;
    $line = $s->getlinebuf;
    return $line if defined $line; # KLUDGE kinda silly/wrong
    return undef if $got == 0;
    return undef;
}

sub getall {
    my ($s) = @_;
    my $all = $s->{linebuf};
    $s->{linebuf} = "";
    while ($s->read(65536)) {
        $all .= $s->{linebuf};
        $s->{linebuf} = "";
    }
    return $all;
}

1;
