use strict;

package SoMat::Sif;
require SoMat::File;
require SoMat::eDAQ::Structurizer;
BEGIN { require 'SoMat/SifConstants.pl'; }

$SoMat::Sif::_datatypemap = {
    SIF_DATATYPE_FLOAT32, ['f', 4],
    SIF_DATATYPE_INT32, ['l', 4],
    SIF_DATATYPE_INT16, ['s', 2],
    SIF_DATATYPE_INT8, ['c', 1],
    SIF_DATATYPE_UNSIGNED32, ['L', 4],
    SIF_DATATYPE_UNSIGNED16, ['S', 2],
    SIF_DATATYPE_UNSIGNED8, ['C', 1]
};

# Note: These are the interfaces from eDAQ/Application/Sif.c that we're
# trying to provide some approximation of the functionality of
# RD_NODE SIF_create(STRING name, BOOL usesRDPtrs)
# PTR SIF_root(void)
# SIF_ChannelList SIF_channel(STRING name, LWORD KeyWord, STRING DataModeType)
# SIF_StandardArea SIF_append(LWORD size)
# void SIF_link(SIF_StandardArea StandardAreaNew, SIF_StandardArea StandardAreaOld)
# void SIF_global(SIF_StandardArea StandardArea, LWORD index)
# void SIF_local(SIF_StandardArea StandardArea, SIF_ChannelList ChannelList, LWORD index)
# void SIF_init_ChannelDef(SIF_ChannelDef ChannelDef)
# void SIF_init_DimensionDef(SIF_DimensionDef DimensionDef)
# void SIF_init_PreDataInfo(SIF_PreDataInfo PreDataInfo)
# void SIF_init_BurstInfo(SIF_BurstInfo BurstInfo)
# LWORD SIF_verify_HdrPtrs(SIF_HeaderRec HeaderRec, LWORD index, BOOL postTest, BOOL *corrupt)
# LWORD SIF_verify_ChanPtrs(SIF_ChannelList ChannelRec, LWORD index, BOOL postTest, BOOL *corrupt)
# LWORD SIF_verify_SAPnext(SIF_StandardArea StandardArea, BOOL postTest, BOOL *corrupt)
# LWORD SIF_verify_SAPself(SIF_StandardArea StandardArea)
# LWORD SIF_sizeof_DataType(LWORD DataType, BOOL resetOnError)
# SIF_StandardArea SIF_traverse(LWORD index)
# SIF_ChannelList SIF_getChannelByName(char *name, BOOL *hasValidFlag)
# SIF_StandardArea SIF_keyStoreTextFile(RD_NODE node, STRING KeyStr, SIF_StandardArea SALink, LWORD *keyDataSize, LWORD *keyDataOfs, BOOL noReset)
# SIF_StandardArea SIF_keyStoreTextMem(char *memImage, LWORD memImageSize, STRING KeyStr, SIF_StandardArea SALink, BOOL noReset)
# void SIF_updatePCMPtrs(LWORD fileCount, LWORD *fsize, LWORD *isRDAHistogram, BOOL *corrupt)
# void SIF_getPadChanNames(LWORD fileCount, LWORD numPadFiles, LWORD *padFileIdxArr, LWORD *numPadChans, LWORD *padChanOffsetArr, BOOL *corrupt)

sub new {
    my $proto = shift;
    my $s = {};
    bless $s, ref($proto) || $proto;
    my $fname = shift;
    #my $mode = SoMat::File::O_RDONLY();
    #my $mode = undef; # KLUDGE testing
    my $mode = shift;
    if (defined $fname) {
        $s->{fname} = $fname;
        if ($fname =~ /^sram:(.*)$/) {
            require SoMat::eDAQ::SRam;
            my $ramdisk = '/dev/ramdisk';
            (my $sram_name = $1) =~ s|/|\\|g;
            my $sram = SoMat::eDAQ::SRam->new($ramdisk);
            return undef unless defined $sram;
            for my $e ($sram->get_dir) {
                if ($e->{name} eq $sram_name) {
                    $s->{fh} = SoMat::File::Offset->new($ramdisk,
                        $mode, $e->{offset}, $e->{size});
                    last;
                }
            }
            undef $sram;
        } else {
            $s->{fh} = SoMat::File::Multi->new($fname, $mode)
                or return undef;
        }
        return undef unless defined $s->{fh};
        $s->{szr} = SoMat::eDAQ::Structurizer->new($s->{fh});
    }
    return $s->_initialize;
}

sub _get_cache_name {
    my ($s) = @_;
    return undef unless exists $s->{fname};
    my $fname = $s->{fname};
    $fname =~ s/\.si[fc]$//i;
    $fname .= '.siy';
    return $fname;
}

sub _spit {
    my ($fh, $v) = @_;

    print $fh 'undef', return unless defined $v;
    my $r = ref($v);
    if (!$r) {
        $v =~ s/([\\\'])/\\$1/g;
        print $fh "'", $v, "'";
    } elsif ($r eq 'HASH') {
        print $fh "{\n";
        my $sep = '';
        foreach my $k (sort keys %$v) {
            print $fh $sep, "'$k'=>";
            _spit($fh, $v->{$k});
            $sep = ",\n";
        }
        print $fh "\n}";
    } elsif ($r eq 'ARRAY') {
        print $fh "[\n";
        my $sep = "";
        foreach my $a (@$v) {
            print $fh $sep;
            _spit($fh, $a);
            $sep = ",\n";
        }
        print $fh "\n]";
    }
}

sub write_cache {
    my ($s) = @_; # KLUDGE should have optional filename,
                  # verbose (include $s->{channels}
    return 1 if $s->{cache_valid};
    return undef unless exists $s->{channel_off} && exists $s->{channel_map};
    # KLUDGE Want to use Storable here if possible
    my $fname = $s->_get_cache_name;
    return undef unless defined $fname;
    my $fh = SoMat::File->new(">$fname");
    return undef unless defined $fh;
    _spit($fh, {'channel_off'=>$s->{channel_off},
        'channel_map'=>$s->{channel_map}
        #, 'channels'=>$s->{channels} # KLUDGE want this to be optional
        });
    print $fh "\n";
    return undef unless $fh->close; # KLUDGE correct check?
    return 1;
}

sub DESTROY {
    my ($s) = @_;
    #close($s->{fh}) if exists $s->{fh};
}

sub SIF_root {
    my ($s) = @_;
    return 0;
}

sub SIF_verify_HdrPtrs {
    my ($s, $pos, $idx) = @_;
    my $szr = $s->{szr};
    $szr->push_type;
    $szr->read('SIF_HeaderRec_OBJ', $pos);
    my $o1 = $szr->get('HdrPtrs', $idx);
    my $o2 = $szr->get('HdrPtrs2', $idx);
    my $o3 = $szr->get('HdrPtrs3', $idx);
    die("SIF_verify_HdrPtrs") if $o1 != $o2 || $o1 != $o3;
    $szr->pop_type;
    return $o1;
}

my $CT_SEQUENTIAL = 1; #Change to 'SEQUENTIAL'
my $CT_HISTOGRAM = 2; #Change to 'HISTOGRAM'
my $CT_HISTANDSEQ = 3; #Change to 'HISTANDSEQ'
my $CT_MESSAGE = 4; # KLUDGE Correct???

#Types of datamodes from ddm_2500.h

#define THIST                           1
#define BURST                           2
#define SEQPV                           3
#define RFLOW                           4
#define PVMATRIX                        5
#define TAL                                     6
#define OTHERSEQ                        7
#define OTHERHIST                       8

#define PCMCIA_THIST            101
#define PCMCIA_BURST            102
#define PCMCIA_SEQPV            103
#define PCMCIA_RFLOW            104
#define PCMCIA_PVMATRIX         105
#define PCMCIA_TAL                      106

#define FM_THIST                        201
#define FM_BURST                        202
#define FM_SEQPV                        203

my $chantype_map = { # Map SA.KeyWord to ChanType (long)
    SIF_KEYWORD_THISTORY, $CT_SEQUENTIAL,
    SIF_KEYWORD_BURST, $CT_SEQUENTIAL,
    SIF_KEYWORD_SEQPV, $CT_SEQUENTIAL,
    SIF_KEYWORD_RAINFLOW, $CT_HISTANDSEQ,
    SIF_KEYWORD_PVMATRIX, $CT_HISTOGRAM,
    SIF_KEYWORD_TAL, $CT_HISTOGRAM,
    7, $CT_SEQUENTIAL, # Hideous KLUDGE for busted header file
    8, $CT_HISTOGRAM, # Hideous KLUDGE for busted header file
    SIF_KEYWORD_THISTORY_RDA, $CT_SEQUENTIAL,
    SIF_KEYWORD_BURST_RDA, $CT_SEQUENTIAL,
    SIF_KEYWORD_SEQPV_RDA, $CT_SEQUENTIAL,
    SIF_KEYWORD_RAINFLOW_RDA, $CT_HISTANDSEQ,
    SIF_KEYWORD_PVMATRIX_RDA, $CT_HISTOGRAM,
    SIF_KEYWORD_TAL_RDA, $CT_HISTOGRAM,
    SIF_KEYWORD_MESSAGE_RDA, $CT_MESSAGE,
    SIF_KEYWORD_THISTORY_RMFDA, $CT_SEQUENTIAL,
    SIF_KEYWORD_BURST_RMFDA, $CT_SEQUENTIAL,
    SIF_KEYWORD_SEQPV_RMFDA, $CT_SEQUENTIAL,
};

$SoMat::Sif::_base_channel = {
    '_offset' => ['SA', 'SAP', 'self'],
    '_size' => ['SA', 'SAP', 'size'],
    '_next' => ['SA', 'SAP', 'next'],
    '_token' => ['SA', 'Token'],
    '_keyword' => ['SA', 'KeyWord'],
    '_chan_def_offset' => ['ChanPtrs', SIF_PTR_CHANNEL_DEF],
    '_data_rec_offset' => ['ChanPtrs', SIF_PTR_DATA_RECORD],
    '_keyword_offset' => ['ChanPtrs', SIF_PTR_KEYWORD_LOCAL],
    'ChanNum' => ['SA', 'LogicalChannel'],
    'ChanName' => ['ChannelName'],
    'RunNum' => ['RunNumber'],
    'DataModeType' => ['DataModeType'],
    'DataModeNum' => ['DataModeNumber']
};

sub _initialize {
    my ($s) = @_;
    return $s if exists $s->{channels};

    my $szr = $s->{szr};
    my $channels = $s->{channels} = [];
    my $channel_off = $s->{channel_off} = [];
    my $channel_map = $s->{channel_map} = {};
    my $root = $s->SIF_root;

    my $o = $s->SIF_verify_HdrPtrs($root, SoMat::Sif::SIF_PTR_CHANNEL_LIST());
    
#print "<p>_initialize before cache", scalar localtime, "<p>\n";
    if (defined (my $cache = $s->_get_cache_name)) {
        my $c = do $cache;
        if (ref($c) eq 'HASH' && exists $c->{channel_off}
            && ref($c->{channel_off}) eq 'ARRAY'
            && @{$c->{channel_off}} > 0) {
            my $ok = 1;
#print "<p>Got cache file $cache\n";
            my $step = $o - $c->{channel_off}->[0];
#print "<p>_initialize: step = $step<p>\n";
            if ($step != 0) {
                local $_;
                map { $_ += $step } @{$c->{channel_off}};
            }
            # Sanity: check first and last channel to be sure the names match
            for my $i (0, @{$c->{channel_off}} - 1) {
                $szr->read('SIF_ChannelList_OBJ',
                    $root + $c->{channel_off}->[$i]);
                # KLUDGE error check from read?
                my $chan_name = $szr->get('ChannelName');
                $ok = 0 unless exists $c->{channel_map}->{$chan_name} 
                    && $c->{channel_map}->{$chan_name} == $i;
                last unless $ok;
                $o = $szr->get('SA', 'SAP', 'next');
            }
#print STDERR "Cache check, last = $last, ok = $ok\n";
            if ($ok) {
                $channel_off = $s->{channel_off} = $c->{channel_off};
                $channel_map = $s->{channel_map} = $c->{channel_map};
                # KLUDGE Do $s->{channels} here too...
                if (exists $c->{channels}) {
                    $channels = $s->{channels} = $c->{channels};
                } else {
                    splice(@$channels, 0, 0, @$channel_off);
                }
                $s->{cache_valid} = ($o == 0);
            }
        }
    }
#print "<p>_initialize after cache", scalar localtime, "<p>\n";
    # If the cache read was successful, we append any new channels.
    # Otherwise we're starting from scratch.

    while ($o) {
        #$szr->read('SIF_ChannelList_OBJ', $root + $o);
        #my $chan_name = $szr->get('ChannelName');
        #$channel_map->{$chan_name} = @$channel_off;
        #push @$channels, $o;
        #$o = $szr->get('SA', 'SAP', 'next');

        # HIDEOUS KLUDGE! equivalent to above
        my $buf = $szr->read(56 + 32, $root + $o); # KLUDGE check errors
        my ($next, $chan_name) = unpack 'x8 L x44 Z32', $buf;

#print "Channel " . scalar @$channel_off . ", chan_name = $chan_name\n";
        $channel_map->{$chan_name} = scalar @$channel_off;
        push @$channel_off, $o;
        push @$channels, $o;
        $o = $next;
    }
#print "<p>_initialize done", scalar localtime, "<p>\n";
    return $s;
}

sub get_channels {
    return $_[0]->{channels};
}

sub get_channel_map {
    return $_[0]->{channel_map};
}

sub get_run_map {
    my ($s) = @_;
    return $s->{run_map} if exists $s->{run_map};
    my %run_map;
    for my $ch (keys %{$s->{channel_map}}) {
        push(@{$run_map{$1}}, $s->{channel_map}->{$ch})
            if $ch =~ /\.RN_(\d+)$/;
    }
    $s->{run_map} = \%run_map;
    return $s->{run_map};
}

sub get_channel_count {
    my ($s) = @_;
    return scalar @{$s->{channels}};
}

sub get_channel {
    my $s = shift;
    my @ch;
    for my $chid (@_) {
        my $val;
        my $idx = -1;
        if (exists $s->{channel_map}->{$chid}) {
            $idx = $s->{channel_map}->{$chid};
            $val = $s->{channels}->[$idx];
        }
        if ($chid =~ /^\d+$/ && $chid >= 0 && $chid < @{$s->{channels}}) {
            $idx = $chid;
            $val = $s->{channels}->[$idx];
        }
        my $ch = SoMat::Sif::Channel->_new($s, $idx, $val);
        push @ch, $ch;
    }
    return $ch[0] if @ch == 1;
    return @ch;
}

sub channel_names {
    my ($s) = @_;
    my @ch;
    my $szr = $s->{szr};

    my $o = $s->SIF_verify_HdrPtrs($s->SIF_root, SIF_PTR_CHANNEL_LIST);
    while ($o) {
        $szr->read('SIF_ChannelList_OBJ', $s->SIF_root + $o);
        push(@ch, $szr->get('ChannelName'));
        $o = $szr->get('SA', 'SAP', 'next'); # Ignore triplicity
    }
    return \@ch;
}

#sub sif_dump {
#    my ($s) = @_;
#    my @ch;
#    my $szr = $s->{szr};
#
#    my $name = "c000";
#
#    my $o = $s->SIF_verify_HdrPtrs($s->SIF_root, SIF_PTR_CHANNEL_LIST);
#    while ($o) {
#        ++$name;
#       my $ch = {};
#        $szr->read('SIF_ChannelList_OBJ', $s->SIF_root + $o);
#       push(@ch, $szr->get('ChannelName'));
#       $szr->dump("${name}.ChannelList");
#my $foo = $szr->get_hash;
#require Data::Dumper;
#print Data::Dumper::Dumper($foo);
#       my $datamode = $szr->get('DataModeType');
#       my $chan_list_kw = $szr->get('SA', 'KeyWord');
#       if ($chan_list_kw == SIF_KEYWORD_THISTORY_RMFDA) {
#print "### SIF_KEYWORD_THISTORY_RMFDA (PCMCIA): $chan_list_kw\n";
#       } elsif ($chan_list_kw == SIF_KEYWORD_THISTORY) {
#print "### SIF_KEYWORD_THISTORY (SRAM): $chan_list_kw\n";
#       } else {
#print "### UNHANDLED: $chan_list_kw\n";
#       }
#
#       my $cho = $szr->get('ChanPtrs', SIF_PTR_CHANNEL_DEF); # Ignore 3ness
#       if ($cho) {
#           $szr->push_type;
#            $szr->read('SIF_ChannelDef_OBJ', 
#                $s->SIF_root + $cho + $szr->size('SIF_StandardArea_OBJ'));
#           $szr->dump("${name}.ChannelDef");
#           $ch->{DataType} = $szr->get('DataType');
#           $ch->{Scale} = $szr->get('Scale');
#           $ch->{Offset} = $szr->get('Offset');
#           my $td = sprintf "%.04d-%.02d-%.02d %.02d:%.02d:%.02d",
#               unpack('S6', $szr->get_raw('TimeDate'));
#           print "${name}.TimeDateXXX = $td\n";
#           if ($datamode eq 'TIMHIS') {
#                $szr->read('SIF_DimensionDef_OBJ');
#               $szr->dump("${name}.DimensionDef.0");
#                $szr->read('SIF_DimensionDef_OBJ');
#               $szr->dump("${name}.DimensionDef.1");
#           }
#           $szr->pop_type;
#       }
#
#       # Dump Channel Keywords
#
#       $cho = $szr->get('ChanPtrs', SIF_PTR_KEYWORD_LOCAL); # Ignore 3ness
#       $szr->push_type;
#       while ($cho) {
#            $szr->read('SIF_StandardArea_OBJ', $s->SIF_root + $cho);
#           my $token = $szr->get('Token');
#           $cho = $szr->get('SAP', 'next');
#           if ($token == SIF_TOKEN_KEYWORD_LOCAL) {
#                $szr->read('SIF_KeyWord_OBJ'); # Data follows SA
#               my $namesize = $szr->get('KeyWordSize');
#               my $valsize = $szr->get('KeyDataSize');
#               my $keyname = $szr->read($namesize);
#               my $keyval = $szr->read($valsize);
#               $keyname =~ s/\000.*//s; #KLUDGE
#               $keyval =~ s/\000.*//s; #KLUDGE
#print "KEYWORD ${name}.$keyname = $keyval\n";
#           }
#       }
#       $szr->pop_type;
#
#       $cho = $szr->get('ChanPtrs', SIF_PTR_DATA_RECORD); # Ignore 3ness
#       $szr->push_type;
#       while ($cho) {
#            $szr->read('SIF_StandardArea_OBJ', $s->SIF_root + $cho);
#           my $token = $szr->get('Token');
#           $szr->dump("###${name}.Data SA");
#           $cho = $szr->get('SAP', 'next');
#           if ($token == SIF_TOKEN_DATA_RECORD) {
#                $szr->read('SIF_PreDataInfo_OBJ'); # Data follows SA
#               $szr->dump("###${name}.PreDataInfo");
#               my $num_points = $szr->get('NumPointsInRec');
#               if ($chan_list_kw == SIF_KEYWORD_THISTORY_RMFDA) {
#                    $szr->read('SIF_RemoteDataPtr_OBJ');
#                    $szr->dump("${name}.SIF_RemoteDataPtr_OBJ");
#                   my $data_off = $szr->get('DataOffset');
#                   print "### DataOffset = $data_off\n";
#                    $szr->read('SIF_FrameMultiplex_OBJ');
#                    $szr->dump("${name}.SIF_FrameMultiplex_OBJ");
#                   my $framesize = $szr->get('FrameSize');
#                   my $frameoff = $szr->get('MuxFrameOffset');
#                   my $framebytes = $szr->get('MuxFrameBytes');
#
#                   my $datatypemap = {
#                       SIF_DATATYPE_FLOAT32, ['f', 4],
#                       SIF_DATATYPE_INT32, ['l', 4],
#                       SIF_DATATYPE_INT16, ['s', 2],
#                       SIF_DATATYPE_INT8, ['c', 1],
#                       SIF_DATATYPE_UNSIGNED32, ['L', 4],
#                       SIF_DATATYPE_UNSIGNED16, ['S', 2],
#                       SIF_DATATYPE_UNSIGNED8, ['C', 1]
#                   };
#
#                   # NOW READ DATA
#
#                  my $i = 0;
#                  my $map = $datatypemap->{$ch->{DataType}};
#                  my ($pack, $samplesize) = @$map;
#                  my $samplesperframe = $framesize / $samplesize;
#                  my $inc = $samplesperframe - 1;
#                  my $packstring = "$pack$samplesperframe";
#print "### DATA TYPE = $ch->{DataType}, pack = $pack, samplesize = $samplesize, packstring = $packstring\n";
#                  $szr->read(0, $s->SIF_root + $data_off + $frameoff);
#                  my @data;
#                  while ($i < 100) {
#                      my $buf = $szr->read($framebytes);
#                      @data[$i .. $i + $inc] = unpack($packstring, $buf);
#                      $i += $samplesperframe;
#                  }
#                  $i = 0;
#                  map { print "+[$i] $_\n"; ++$i } @data;
#
#                   $szr->read('SIF_StandardArea_OBJ', $s->SIF_root + $cho);
#               }
#           }
#
#       }
#       $szr->pop_type;
#
#        #$o = $s->SIF_verify_SAPnext(Channel->SA);
#       my $o1 = $szr->get('SA', 'SAP', 'next');
#       my $o2 = $szr->get('SA', 'SAP2', 'next');
#       my $o3 = $szr->get('SA', 'SAP3', 'next');
#        die("SIF_SAP_next") if $o1 != $o2 || $o1 != $o3;
#       $o = $o1;
#    }
#    return \@ch;
#}

package SoMat::Sif::Channel;

sub _new {
    my ($class, $sif, $idx, $val) = @_;
    return undef unless defined $val;
    my $s = { '_Sif' => $sif, '_Values' => $val };
    bless $s, ref($class) || $class;
    $s->_init_values($idx, $val);
    return $s;
}

sub _init_values {
    my ($s, $idx, $o) = @_;
    return if ref($o) eq 'HASH';

    my $sif = $s->{_Sif};
    my $szr = $sif->{szr};

    $szr->push_type;
    $szr->read('SIF_ChannelList_OBJ', $sif->SIF_root + $o);
    my %ch;
    while (my ($key, $val) = each %$SoMat::Sif::_base_channel) {
        $ch{$key} = $szr->get(@$val);
    }
$ch{ChanType} = "Mapping undefined ($ch{_keyword})"; # TESTING KLUDGE
    $ch{ChanType} = $chantype_map->{$ch{_keyword}}
        if exists $chantype_map->{$ch{_keyword}};
    if ($ch{ChanName} =~ /^(\w+)\@(\w+)\.RN_\d+$/) {
        $ch{DataModeName} = $1;
        $ch{BaseChanName} = $2;
    }
    $szr->pop_type;
    $sif->{channels}->[$idx] = $s->{_Values} = \%ch;
}

sub Get {
    my ($s, $property, $index, $metadata, $reference) = @_;
    if (ref $property) {
        my $i = 0;
        my @ret;
        foreach (@$property) {
            push @ret, $s->Get($_, $index->[$i], $metadata->[$i], $reference);
            $i++;
        }
        return @ret;
    }
    return $s->{_Values}->{$property} if exists $s->{_Values}->{$property};
    return $s->_Get($property, $index, $metadata, $reference);
}

sub GetKeyword {
    my ($s, $keyword) = @_;
    $s->_get_keywords unless exists $s->{_Keywords};
    my $kw = $s->{_Keywords};
    return $kw unless defined $keyword;
    return (exists $kw->{$keyword}) ? $kw->{$keyword} : undef;
}

sub _get_keywords {
    my ($s) = @_;
    my $off = $s->{_Values}->{_keyword_offset};
    my $kw = $s->{_Keywords} = {};
    return undef unless $off;
    my $sif = $s->{_Sif};
    my $szr = $sif->{szr};
    my $v = $s->{_Values};
    $szr->push_type;
    while ($off) {
        $szr->read('SIF_StandardArea_OBJ', $sif->SIF_root + $off);
        my $token = $szr->get('Token');
        $off = $szr->get('SAP', 'next');
        if ($token == SoMat::Sif::SIF_TOKEN_KEYWORD_LOCAL()) {
            $szr->read('SIF_KeyWord_OBJ'); # Data follows SA
            my $namesize = $szr->get('KeyWordSize');
            my $valsize = $szr->get('KeyDataSize');
            my $keyname = $szr->read($namesize);
            my $keyval = $szr->read($valsize);
            $keyname =~ s/\000.*//s; #KLUDGE
            $keyval =~ s/\000.*//s; #KLUDGE
            $kw->{$keyname} = $keyval;
        }
    }
    $szr->pop_type;
    return 1;
}

sub _Get {
    my ($s, $property, $index, $metadata, $reference) = @_;
    my $sub = "__$property";
    return undef unless $s->can($sub);
    return $s->$sub($property, $index, $metadata, $reference);
}

my $channel_def = {
    'NumDims' => ['NoDims'],
    'Scale' => ['Scale'],
    'Offset' => ['Offset'],
    'DataType' => ['DataType'],
};

my $dimen_def_0 = {
    'NumPoints' => ['NumPointsOrBins'],
    'BaseTime' => ['Base'],
    'SampleInc' => ['Increment'],
};

my $dimen_def_1 = {
};

my $dimen_map = {
    'NumPoints' => ['NumPointsOrBins'],
    'BaseTime' => ['Base'],
    'SampleInc' => ['Increment'],
    'Label' => ['Label'],
    'Units' => ['Units'],
    'BinType' => ['BinType'],
    '_bin_def_offset' => ['BinDefPtr'],
    'AxisLabel.Dim' => ['Label'],
    'AxisUnits.Dim' => ['Units'],
    'NumBins.Dim' => ['NumPointsOrBins'],
};

sub _ChannelDefField {
    my ($s, $property, $index, $metadata, $reference) = @_;
    #'_chan_def_offset' => ['ChanPtrs', SIF_PTR_CHANNEL_DEF],
    my $off = $s->{_Values}->{_chan_def_offset};
    return undef unless $off;
    my $sif = $s->{_Sif};
    my $szr = $sif->{szr};
    my $v = $s->{_Values};
    $szr->push_type;
    $szr->read('SIF_ChannelDef_OBJ', 
        $sif->SIF_root + $off + $szr->size('SIF_StandardArea_OBJ'));
    while (my ($key, $val) = each %$channel_def) {
        $v->{$key} = $szr->get(@$val);
#print "_ChannelDefField: $key => $v->{$key}\n";
    }
    $v->{SampleBits} = $v->{DataType} & 0xFF;
    my @timedate = unpack('S6', $szr->get_raw('TimeDate'));
    $v->{TestDate} = sprintf("%.04d-%.02d-%.02d", @timedate);
    $v->{TestTime} = sprintf("%.02d:%.02d:%.02d", @timedate[3..5]);
    $v->{TestDateTime} = "$v->{TestDate}T$v->{TestTime}";
    #if ($v->{DataModeType} eq 'TIMHIS') {
    #}
    for my $i (1 .. $v->{NumDims}) {
        $szr->read('SIF_DimensionDef_OBJ');
        my %dim;
        while (my ($key, $val) = each %$dimen_map) {
            if ($key =~ /\.Dim$/) {
                $v->{"$key$i"} = $szr->get(@$val);
            } else {
                $dim{$key} = $szr->get(@$val);
            }
        }
        push(@{$s->{_Dimensions}}, \%dim);
    }
    if ($v->{DataModeType} eq 'MESSAGE') {
        $szr->read('SIF_MessageDef_OBJ');
        # KLUDGE
        for my $f (qw(FrameCount MessageCount)) {
            $v->{$f} = $szr->get($f);
        }
    }
    if ($v->{NumDims}) {
        my $dim0 = $s->{_Dimensions}->[0];
        $v->{NumPoints} = $dim0->{NumPoints};
        $v->{BaseTime} = $dim0->{BaseTime};
        $v->{SampleInc} = $dim0->{SampleInc};
        $v->{SampleRate} =
            ($v->{ChanType} == int($CT_HISTANDSEQ) || $dim0->{SampleInc} == 0)
            ? 1 : sprintf('%.4g', 1.0 / $dim0->{SampleInc});
    }
    $szr->pop_type;
    return (exists $v->{$property}) ? $v->{$property} : undef;
}

# Values from $channel_def
*__NumDims = \&_ChannelDefField;
*__Scale = \&_ChannelDefField;
*__Offset = \&_ChannelDefField;
*__DataType = \&_ChannelDefField;

# Other unconditional values
*__SampleBits = \&_ChannelDefField;
*__SampleRate = \&_ChannelDefField;
*__TestDate = \&_ChannelDefField;
*__TestDateTime = \&_ChannelDefField;
*__TestTime = \&_ChannelDefField;

# Values if NumDims
*__NumPoints = \&_ChannelDefField;
*__BaseTime = \&_ChannelDefField;
*__SampleInc = \&_ChannelDefField;
*__SampleRate = \&_ChannelDefField;
*__BinType = \&_ChannelDefField;

{ # KLUDGE!!! AUTOLOAD!!!
no strict 'refs';
*{'__AxisUnits.Dim1'} = \&_ChannelDefField;
*{'__AxisLabel.Dim1'} = \&_ChannelDefField;
*{'__AxisUnits.Dim2'} = \&_ChannelDefField;
*{'__AxisLabel.Dim2'} = \&_ChannelDefField;
*{'__AxisUnits.Dim3'} = \&_ChannelDefField;
*{'__AxisLabel.Dim3'} = \&_ChannelDefField;
*{'__AxisUnits.Dim4'} = \&_ChannelDefField;
*{'__AxisLabel.Dim4'} = \&_ChannelDefField;
*{'__AxisUnits.Dim5'} = \&_ChannelDefField;
*{'__AxisLabel.Dim5'} = \&_ChannelDefField;
*{'__AxisUnits.Dim6'} = \&_ChannelDefField;
*{'__AxisLabel.Dim6'} = \&_ChannelDefField;
*{'__AxisUnits.Dim7'} = \&_ChannelDefField;
*{'__AxisLabel.Dim7'} = \&_ChannelDefField;
*{'__AxisUnits.Dim8'} = \&_ChannelDefField;
*{'__AxisLabel.Dim8'} = \&_ChannelDefField;
}

sub __PhysicalChan {
    my ($s, $property, $index, $metadata, $reference) = @_;
    return $s->GetKeyword('DM_PhysicalChan');
}

sub GetData {
    my ($s, $offset, $count) = @_;
    #'_data_rec_offset' => ['ChanPtrs', SIF_PTR_DATA_RECORD],
    my $off = $s->{_Values}->{_data_rec_offset};
    return undef unless $off;
    my $sif = $s->{_Sif};
    my $szr = $sif->{szr};
    my @data;

    $offset = 0 unless defined $offset;

#my $name = $s->Get('ChanName');
    $szr->push_type;
    while ($off) {
        $szr->read('SIF_StandardArea_OBJ', $sif->SIF_root + $off);
        my $token = $szr->get('Token');
#$szr->dump("###${name}.Data SA");
        $off = $szr->get('SAP', 'next');
        if ($token == SoMat::Sif::SIF_TOKEN_DATA_RECORD()) {
            $szr->read('SIF_PreDataInfo_OBJ'); # Data follows SA
#$szr->dump("###${name}.PreDataInfo");
            my $num_points = $szr->get('NumPointsInRec');
            $count = $num_points unless defined $count;
            $count = $num_points - $offset if $count + $offset > $num_points;
            return undef if $count < 0;
#print "num_points = $num_points\n";
            my $keyword = $s->Get('_keyword');
            if ($keyword == SoMat::Sif::SIF_KEYWORD_THISTORY_RMFDA()) {
                local $_;
                $szr->read('SIF_RemoteDataPtr_OBJ');
#$szr->dump("${name}.SIF_RemoteDataPtr_OBJ");
                my $data_off = $szr->get('DataOffset');
#print "### DataOffset = $data_off\n";
                $szr->read('SIF_FrameMultiplex_OBJ');
#$szr->dump("${name}.SIF_FrameMultiplex_OBJ");
                my $framesize = $szr->get('FrameSize');
                my $frameoff = $szr->get('MuxFrameOffset');
                my $framebytes = $szr->get('MuxFrameBytes');

                # NOW READ DATA

                my $i = 0;
                my ($datatype, $eng_scale, $eng_offset)
                    = $s->Get(['DataType', 'Scale', 'Offset']);
                my $map = $SoMat::Sif::_datatypemap->{$datatype};
                my ($pack, $samplesize) = @$map;
                my $samplesperframe = $framesize / $samplesize;
                my $inc = $samplesperframe - 1;
                my $packstring = "$pack$samplesperframe";
                my $skip_samples = $offset % $samplesperframe;
                my $skip_frames = ($offset - $skip_samples) / $samplesperframe;
                $szr->read(0, $sif->SIF_root + $data_off
                    + $skip_frames * $framebytes);
                my $frames = int(
                    ($skip_samples + $count + $inc) / $samplesperframe);
                my $buf = $szr->read($framebytes * $frames);
                # KLUDGE error check here
                my $pos = $frameoff;
                while ($frames--) {
                    push @data,
                        unpack($packstring, substr($buf, $pos, $framesize));
                    $pos += $framebytes;
                }
                splice @data, 0, $skip_samples;
                splice(@data, $count, scalar(@data) - $count);

                map { $_ = $_ * $eng_scale + $eng_offset } @data
                    if $eng_scale != 1.0 || $eng_offset != 0.0;
                last;
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_MESSAGE_RDA()) {
                my $hdr_size = $szr->size('SIF_MsgFrameHeader_OBJ');
                $szr->read('SIF_RemoteDataPtr_OBJ');
                my $data_off = $szr->get('DataOffset');
                #$szr->read('SIF_FrameMultiplex_OBJ');

                my $bytes = $num_points; 
                my $t_lo_factor = 1 / $s->Get('SampleRate');
                my $t_hi_factor = (1.0 << 32) * $t_lo_factor;
                my $bo = 0;
                my $mo = 0;
                if ($offset > 0 && exists $s->{_msg_offset_cache}) {
                    while (my ($tmo, $tbo) = each %{$s->{_msg_offset_cache}}) {
                        if ($tmo <= $offset && $tmo > $mo) {
                            $bo = $tbo;
                            $mo = $tmo;
                        }
                    }
                    $bytes -= $bo;
                }
                $offset -= $mo;
                $szr->read(0, $sif->SIF_root + $data_off + $bo);
                my $o = 0;
                my $buf = '';
                my $bufsiz = 40000;
                my $msg = '';
                my $time = undef;
                while ($count > 0 && $bytes > 0) {
                    if ($o + $hdr_size > length($buf)) {
                        substr($buf, 0, $o, '');
                        $bo += $o;
                        $o = 0;
                        $buf .= $szr->read($bufsiz);
                        last if $o + $hdr_size > length($buf);
                    }
                    my ($packetsize, $timestamp_hi, $timestamp_lo)
                        = unpack('SLL', substr($buf, $o, $hdr_size));
                    my $continued = ($packetsize & 0x8000);
                    $packetsize &= 0x7FFF;
                    # KLUDGE the duplication of code here with above is bad
                    if ($o + $packetsize > length($buf)) {
                        substr($buf, 0, $o, '');
                        $bo += $o;
                        $o = 0;
                        $buf .= $szr->read($bufsiz);
                        last if $o + $packetsize > length($buf);
                    }
                    if ($offset > 0) {
                        --$offset;
                    } else {
                        unless (defined $time) {
                            $time = $timestamp_hi * $t_hi_factor
                                + $timestamp_lo * $t_lo_factor
                        }
                        $msg .= substr($buf, $o + $hdr_size,
                            $packetsize - $hdr_size);
                        if (!$continued) {
                            push @data, $time, $msg;
                            $msg = '';
                            $time = undef;
                        }
                        --$count;
                    }
                    $o += $packetsize;
                    ++$mo;
                    $bytes -= $packetsize;
                }
                $s->{_msg_offset_cache}->{$mo} = $bo + $o;
                last; # KLUDGE???
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_THISTORY()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_THISTORY\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_BURST()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_BURST\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_SEQPV()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_SEQPV\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_RAINFLOW()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_RAINFLOW\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_PVMATRIX()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_PVMATRIX\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_TAL()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_TAL\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_THISTORY_RDA()) {
# NOT EDAQ???
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_THISTORY_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_BURST_RDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_BURST_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_SEQPV_RDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_SEQPV_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_RAINFLOW_RDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_RAINFLOW_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_PVMATRIX_RDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_PVMATRIX_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_TAL_RDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_TAL_RDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_BURST_RMFDA()) {
# NOT EDAQ???
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_BURST_RMFDA\n";
            } elsif ($keyword == SoMat::Sif::SIF_KEYWORD_SEQPV_RMFDA()) {
                # eDAQ
                warn "GetData: Unsupported DATA_RECORD: KEYWORD_SEQPV_RMFDA\n";
            } else {
                warn "GetData: Unsupported DATA_RECORD: $keyword\n";
            }
        } else {
            warn "GetData: unknown token $token;";
        }
    }
    $szr->pop_type;
    return undef unless @data;
    return \@data;
}

1;

