use strict;

package SoMat::SifExtract;

require SoMat::File;

#use Carp qw(confess);
#BEGIN { $SIG{__DIE__} = sub { require Carp; Carp::confess(@_) } }

# The SIF file definition requires that the following values never change
# "use constant" enables constant folding and other perl optimizations

# Sizes
use constant sz_BurstInfo => 16;
use constant sz_ChannelDef => 48;
use constant sz_ChannelList => 240;
use constant sz_DimensionDef => 92;
use constant sz_FrameMultiplex => 60;
use constant sz_HeaderRec => 128;
use constant sz_KeyWord => 8;
use constant sz_PreDataInfo => 20;
use constant sz_RemoteDataPtr => 24;
use constant sz_StandardArea => 56;

# Keywords
use constant kw_THISTORY => 1;
use constant kw_BURST => 2;
use constant kw_SEQPV => 3;
use constant kw_RAINFLOW => 4;
use constant kw_PVMATRIX => 5;
use constant kw_TAL => 6;
use constant kw_THISTORY_RDA => 101;
use constant kw_BURST_RDA => 102;
use constant kw_SEQPV_RDA => 103;
use constant kw_RAINFLOW_RDA => 104;
use constant kw_PVMATRIX_RDA => 105;
use constant kw_TAL_RDA => 106;
use constant kw_MESSAGE_RDA => 150;
use constant kw_THISTORY_RMFDA => 201;
use constant kw_BURST_RMFDA => 202;
use constant kw_SEQPV_RMFDA => 203;

# Tokens
use constant tok_HEADER_RECORD => 1;
use constant tok_CHANNEL_LIST => 2;
use constant tok_CHANNEL_DEF => 3;
use constant tok_DATA_RECORD => 4;
use constant tok_BIN_DEF => 5;
use constant tok_KEYWORD_GLOBAL => 7;
use constant tok_KEYWORD_LOCAL => 8;
use constant tok_STACK_RECORD => 9;
use constant tok_HIDDEN_RECORD => 0xFFFF;

# Ptrs
use constant ptr_CHANNEL_LIST => 0;
use constant ptr_KEYWORD_GLOBAL => 3;
use constant ptr_RAMDISK_REMOTE => 7;
use constant ptr_CHANNEL_DEF => 0;
use constant ptr_DATA_RECORD => 1;
use constant ptr_KEYWORD_LOCAL => 2;
use constant ptr_DATA_RECORD_EOF => 3;
use constant ptr_TIME_CHANNEL => 4;
use constant ptr_STACK_RECORD => 5;

# SifExtract internal stuff
use constant DecomposeChunk => 'nN';
sub _n_base_size () { 16 } # KLUDGE use constant 5.5.3 fails for /^_/
sub _n_ref_size () { 12 }

# The methods below are used to create a map of the various parts of a SIF
# file. Once maps have been created for each input SIF file, the main
# remapping facilities later on in the code can be used to write a new
# SIF file from selected parts of the input SIF files.

sub new_sif {
    my $parent = shift;
    my $s = bless {}, ref($parent);
    $s->{current_ch} = 0;
    $s->{samplesize} = [];
    return $s->_initialize_sif($parent->{next_id}++, @_);
}

sub get {
    my ($s, $name) = @_;
    return $s->{$name};
}

sub _initialize_sif {
    my ($s, $id, $fname) = @_;
    $s->{id} = $id;
    $s->{fname} = $fname;
    if ($fname) {
        $s->{fh} = $s->open_sif_file($fname) or return undef;
    }
    $s->{map} = {};
    return $s;
}

sub open_sif_file {
    my ($s, $fname, $mode) = @_;
    return undef unless defined $fname;
    if ($fname =~ /^sram:(.*)$/) {
        (my $sram_name = $1) =~ s|/|\\|g;
        require SoMat::eDAQ::SRam;
        my $ramdisk = '/dev/ramdisk';
        my $sram = SoMat::eDAQ::SRam->new($ramdisk);
        return undef unless defined $sram;
        for my $e ($sram->get_dir) {
            if ($e->{name} eq $sram_name) {
                return SoMat::File::Offset->new($ramdisk, $mode,
                    $e->{offset}, $e->{size});
            }
        }
        return undef;
    }
    return SoMat::File::Multi->new($fname, $mode);
}

sub _read {
    my ($s, $size, $offset) = @_;
    return ($s->{packed} = "\x00" x $size) unless $s->{fh};
    # KLUDGE die on errors rather than return undef?
    if (defined $offset) {
        $s->{fh}->seek($offset, 0) == $offset or return undef;
    }
    return undef unless $s->{fh}->read($s->{packed}, $size) == $size;
    return $s->{packed};
}

sub new_node {
    my ($s, $offset, $size) = @_;
    my $node = _n_new();
    _n_set_sif_id($node, $s->{id});
    _n_set_offset($node, $offset);
    _n_set_size($node, $size);
    $s->{map}->{$offset} = $node;
}

sub add_node {
    my ($s, $offset) = @_;
    return if exists $s->{map}->{$offset};
    my $node = _n_new();
    _n_set_sif_id($node, $s->{id});
    _n_set_offset($node, $offset);
    _n_set_logicalchannel($node, $s->{current_ch});

    my @r;
    # It's a StandardArea:
    #   SAP.self   SAP.size   SAP.next   SAP.prev   (V4)
    #  SAP2.self  SAP2.size  SAP2.next  SAP2.prev   (V4)
    #  SAP3.self  SAP3.size  SAP3.next  SAP3.prev   (V4)
    #  LogicalChannel (V) Token (v)  KeyWord (v)
    # WARNING - For efficiency, NO triplicity checks are done!

    my $sa = $s->_read(sz_StandardArea, $offset);
    my ($self, $size, $next, $prev, $tok, $kw) = unpack('V4x36vv', $sa);
    _n_add_3ref($node, $s->{id}, $self, 0, 16) if $self;
    _n_add_3ref($node, $s->{id}, $next, 8, 16) if $next;
    _n_add_3ref($node, $s->{id}, $prev, 12, 16) if $prev;
    _n_set_size($node, $size); # SA SAP.size

    if ($tok == tok_CHANNEL_LIST) {
        #my ($chname, @chptrs) = unpack('Z32x16V8', $s->_read(sz_ChannelList));
        my @ChanPtrs = unpack('x48V8', $s->_read(sz_ChannelList));
        my $o = 104; # ChanPtrs[0] offset
        for my $ptr (@ChanPtrs) {
            if ($ptr) {
                _n_add_3ref($node, $s->{id}, $ptr, $o, 32);
                push @r, $ptr;
            }
            $o += 4;
        }
        # We ignore ExpansionPtrs; useless due to broken SIF translator
        $next = 0; # Only do this channel
    } elsif ($tok == tok_CHANNEL_DEF) {
        my ($NoDims, $DataType) = unpack('vv', $s->_read(sz_ChannelDef));
        my $o = sz_StandardArea + sz_ChannelDef;
        $s->{samplesize}->[$s->{current_ch}] = ($DataType & 0xFF) / 8;

        for my $i (1 .. $NoDims) {
            my $ptr = unpack('x80V', $s->_read(sz_DimensionDef)); # BinDefPtr
            if ($ptr) {
                # BinDefPtr, BinDefPtr2, BinDefPtr3
                _n_add_3ref($node, $s->{id}, $ptr, $o + 80, 4);
                push @r, $ptr;
            }
            $o += sz_DimensionDef;
        }
    } elsif ($tok == tok_DATA_RECORD || $tok == tok_STACK_RECORD) {

        # SA = StandardArea, PDI = PreDataInfo
        #
        # THISTORY => [SA, PDI, data]
        # SEQPV    => [SA, PDI, data]
        # TAL      => [SA, PDI, data]
        # RAINFLOW => [SA, PDI, data]
        # RAINFLOW => [SA, PDI, data]
        # PVMATRIX => [SA, PDI, data]
        #
        # BURST     => [SA, PDI, BurstInfo, data]
        # BURST_RDA => [SA, PDI, BurstInfo, RemoteDataPtr] ... [RDATA]
        #
        # THISTORY_RDA => [SA, PDI, RemoteDataPtr] ... [RDATA]
        # SEQPV_RDA    => [SA, PDI, RemoteDataPtr] ... [RDATA]
        # TAL_RDA      => [SA, PDI, RemoteDataPtr] ... [RDATA]
        # RAINFLOW_RDA => [SA, PDI, RemoteDataPtr] ... [RDATA]
        # PVMATRIX_RDA => [SA, PDI, RemoteDataPtr] ... [RDATA]
        # MESSAGE_RDA  => [SA, PDI, RemoteDataPtr] ... [RDATA]
        #
        # THISTORY_RMFDA => [SA, PDI, RemoteDataPtr, FrameMultiplex] ... [RDATA]
        # SEQPV_RMFDA    => [SA, PDI, RemoteDataPtr, FrameMultiplex] ... [RDATA]

        if ($kw >= kw_THISTORY_RDA) {
            my $samplecount = unpack('x4V', $s->_read(sz_PreDataInfo));
            my $samplesize = $s->{samplesize}->[$s->{current_ch}];
            my $size = $samplecount * $samplesize;
            my $o = sz_StandardArea + sz_PreDataInfo;

            if ($kw == kw_BURST_RDA || $kw == kw_BURST_RMFDA) {
                $s->_read(sz_BurstInfo);
                $o += sz_BurstInfo;
            }
            my $ptr = unpack('V', $s->_read(sz_RemoteDataPtr)); # DataOffset
            if ($kw == kw_THISTORY_RMFDA || $kw == kw_SEQPV_RMFDA) {
                my ($FrameSize, $MuxFrameBytes)
                    = unpack('Vx20V', $s->_read(sz_FrameMultiplex));
                # Do NOT increment $o : FrameMultiplex is _after_ RemoteDataPtr
                my $samples_per_frame = $FrameSize / $samplesize;
                my $frames = $samplecount / $samples_per_frame;
                $frames = int(++$frames) if int($frames) < $frames; # ceil()
                $size = $MuxFrameBytes * $frames;
            } elsif ($kw == kw_PVMATRIX_RDA) {
                $size += 3 * $samplesize; # KLUDGE - Why 3?
            } elsif ($kw == kw_TAL_RDA) {
                $size += 1 * $samplesize; # KLUDGE - Why 1?
            } elsif ($kw == kw_RAINFLOW_RDA) {
                if ($tok == tok_DATA_RECORD) {
                    $size += 1 * $samplesize; # KLUDGE - Why 1?
                } elsif ($tok == tok_STACK_RECORD) {
                    $size = 257 * $samplesize; # KLUDGE - Why 257?
                }
            }
            if ($ptr) {
                # DataOffset, DataOffset2, DataOffset3
                _n_add_3ref($node, $s->{id}, $ptr, $o, 4);
                my $dn = _n_new();
                _n_set_sif_id($dn, $s->{id});
                _n_set_offset($dn, $ptr);
                _n_set_size($dn, $size);
                _n_set_is_data($dn, 1);
                $s->{map}->{$ptr} = $dn;
            }
        }
    } elsif ($tok == tok_HEADER_RECORD) {
        my @HdrPtrs = unpack('x32V8', $s->_read(sz_HeaderRec));
        my $o = 88; # sz_StandardArea + offset HeaderRec.HdrPtrs_0
        for my $ptr (@HdrPtrs) {
            _n_add_3ref($node, $s->{id}, $ptr, $o, 32);
            push(@r, $ptr) if $ptr;
            $o += 4;
        }
    }
    $s->{map}->{$offset} = $node;
    push(@r, $next) if $next;
    return @r;
}

sub add_nodes {
    my ($s, @offs) = @_;
    # We handle the stack ourselves to eliminate recursion limit errors
    # and save memory
    while (@offs) {
        my $o = shift @offs;
        unshift @offs, $s->add_node($o);
    }
}

sub map {
    my ($s) = @_;
    my $fh = $s->{fh};
    my ($next, $packed, %channel_list, @channels, $ch);

    $fh->seek(0) == 0 or return undef;
    $fh->read($packed, sz_StandardArea + sz_HeaderRec)
        == sz_StandardArea + sz_HeaderRec or return undef;
    $s->{hdrptrs} = [ unpack('x88V8', $packed) ];

    my $lchan = 1;
    my $o = $s->{hdrptrs}->[ptr_CHANNEL_LIST];
    my %ch_offs;
    while ($o) {
        return undef if $ch_offs{$o}++; # Pointer infinite loop detector
        $fh->seek($o) == $o or return undef;
        $fh->read($packed, sz_StandardArea + sz_ChannelList)
            == sz_StandardArea + sz_ChannelList or return undef;
        ($next, $ch) = unpack('x8Vx44Z32', $packed);
        $channel_list{$ch} = pack('VV', $o, $lchan++);
        push @channels, $ch;
        $o = $next;
    }
    $s->{channel_list} = \%channel_list;
    $s->{channels} = \@channels;
    return 1;
}

sub map_global_keywords {
    my ($s) = @_;
    return if exists $s->{global_keywords};
    my $fh = $s->{fh};
    my ($packed, %keyword_list, @keywords);

    my $o = $s->{hdrptrs}->[ptr_KEYWORD_GLOBAL];
    while ($o) {
        $fh->seek($o);
        $fh->read($packed, sz_StandardArea + sz_KeyWord);
        my ($next, $name_size, $val_size) = unpack('x8Vx44VV', $packed);
        $fh->read($packed, $name_size);
        my $kw_name = unpack('Z*', $packed);
        $keyword_list{$kw_name} = $o;
        push @keywords, $o;
        $o = $next;
    }
    $s->{global_keyword_list} = \%keyword_list;
    $s->{global_keywords} = \@keywords;
}

# The following stuff is the meat of the SIF chunk remapping facility

sub new {
    my $proto = shift;
    my $s = bless {}, ref($proto) || $proto;
    $s->{next_id} = 0;
    $s->_initialize_sif($s->{next_id}++, @_);
    $s->{sources} = [];
    $s->{off} = 0;
    $s->{links} = {};
    $s->{remap} = {};
    # Any keys of $s that are integers are output blocks for that offset
    $s->_create_header_rec;
    return $s;
}

sub _create_header_rec {
    my ($s) = @_;

    my $nsize = sz_StandardArea + sz_HeaderRec;
    $s->new_node(0, $nsize);

    my $r = "\x00" x $nsize;
    my $psize = pack('V', $nsize);
    substr($r, 4,  4, $psize); # SA.SAP.size
    substr($r, 20, 4, $psize); # SA.SAP2.size
    substr($r, 36, 4, $psize); # SA.SAP2.size
    substr($r, 52, 2, pack('v', tok_HEADER_RECORD)); # SA.Token
    substr($r, 56, 32, pack('Z32', 'SoMateDAQPCMDataFile-v1.1')); # IDString
    $s->{0} = $r;

    $s->remap_ref($s->chunk(0));
    # All links are already zero
}

sub chunk {
    my ($s, $off, $id) = @_;
    $id = $s->{id} unless defined $id;
    return pack('nN', $id, $off);
}

sub add_sif {
    my ($s, $fname) = @_;
    my $sif = $s->new_sif($fname);
    die "Can't add $fname: $@" unless defined $sif;
    return undef unless $sif->map;
    $s->{sources}->[$sif->{id}] = $sif;
    return $sif;
}

sub _add_link {
    my ($s, $idx, $sif, $offset) = @_;
    my $chunk = $s->chunk($offset, $sif->{id});
    push @{$s->{links}->{$idx}}, $chunk;
    return $chunk;
}

sub add_channel {
    my ($s, $sif, $chname) = @_;
    my $packed = $sif->{channel_list}->{$chname};
    return undef unless $packed; # KLUDGE
    my ($offset, $lchan) = unpack('VV', $packed);
    $sif->{current_ch} = $lchan;
    $sif->add_nodes($offset);
    return $s->_add_link(ptr_CHANNEL_LIST, $sif, $offset);
}

sub add_global_keyword {
    my ($s, $sif, $kw_name) = @_;
    $sif->map_global_keywords;
    my $offset = $sif->{global_keyword_list}->{$kw_name} or return undef;
    $sif->add_nodes($offset);
    return $s->_add_link(ptr_KEYWORD_GLOBAL, $sif, $offset);
}

sub add_all_global_keywords {
    my ($s, $sif) = @_;
    $sif->map_global_keywords;
    for my $offset (@{$sif->{global_keywords}}) {
        $sif->add_nodes($offset);
        $s->_add_link(ptr_KEYWORD_GLOBAL, $sif, $offset);
    }
}

sub link_header_list {
    my ($s, $hdr_idx, $list) = @_;
    $list = [] unless defined $list;
    my $czero = $s->chunk(0);
    my $hdr = $s->get_chunk_entry($czero);
    my $head = (@$list) ? $list->[0] : $czero;

    # Link the HeaderRec HdrPtrs*
    my $o = 88 + 4 * $hdr_idx;
    my ($hid, $hoff) = unpack(DecomposeChunk, $head);
    _n_add_3ref($$hdr, $hid, $hoff, $o, 32);

    for (my $i = 0; $i < @$list; $i++) {
        my $prev = ($i == 0) ? $czero : $list->[$i - 1];
        my $self = $list->[$i];
        my $next = ($i >= @$list - 1) ? $czero : $list->[$i + 1];
        $s->remap_ref($self);
        my $node = $s->get_chunk_entry($self);
        if ($hdr_idx == ptr_CHANNEL_LIST) {
            $s->{LogicalChannel} = $i + 1;
            _n_set_logicalchannel($$node, $i + 1);
        }
        $s->_n_put_sa_ref($$node, $czero, $self, $czero);
        $s->remap_ref_recursive($self);
        $s->_n_put_sa_ref($$node, $prev, $self, $next);
        $s->{LogicalChannel} = undef;
    }
}

sub get_chunk_entry {
    my ($s, $chunk) = @_;
    if (defined (my $remap = $s->{remap}->{$chunk})) {
        return \($s->{map}->{$remap}) if defined $s->{map}->{$remap};
    }
    my ($id, $off) = unpack(DecomposeChunk, $chunk);
    my $sif = ($s->{id} == $id ? $s : $s->{sources}->[$id]) or return undef;
    return \($sif->{map}->{$off}) if defined $sif->{map}->{$off};
    return undef;
}

sub remap_ref {
    my ($s, $chunk, $recurse) = @_;

    my $er = $s->get_chunk_entry($chunk);
    return if $recurse && defined $er && _n_remapped($$er);
    unless (exists $s->{remap}->{$chunk}) {
        my ($cid, $coff) = unpack(DecomposeChunk, $chunk);
        if ($s->{id} == $cid && exists $s->{map}->{$coff}) {
            $s->{remap}->{$chunk} = $coff;
            $s->{off} += length $s->{$coff};
        } else {
            my $e = $$er;
            _n_set_logicalchannel($e, $s->{LogicalChannel})
                if defined $s->{LogicalChannel};
            $s->{map}->{$s->{off}} = $e;
            $s->{remap}->{$chunk} = $s->{off};
            $s->{off} += _n_size($e);
        }
    }
    return unless $recurse;
    _n_set_remapped($$er, 1);
    return grep { $_ && $_ ne $chunk } _n_refs_values($$er);
}

sub remap_ref_recursive {
    my ($s, @refs) = @_;
    while (@refs) {
        my $r = shift @refs;
        unshift @refs, $s->remap_ref($r, 1);
    }
}

sub relink {
    my ($s) = @_;

    for my $i (sort keys %{$s->{links}}) {
        $s->link_header_list($i, $s->{links}->{$i});
    }
    for my $o (sort {$a <=> $b} keys %{$s->{map}}) {
        $s->remap_ref_recursive($s->chunk($o));
    }
}

sub emit {
    # All relinking must be done by the time this is called
    # i.e. all VALUES of $s->{remap} must be integers
    my ($s, $fh) = @_;
    my $map = $s->{map};
    my @source = @{$s->{sources}};
    $source[$s->{id}] = $s;
    my $cout = 0;
    use constant OutSz => 65536;

    for my $chunk (sort {$a <=> $b} keys %$map) {
        my $e = $map->{$chunk};
        my $left = _n_size($e);
        my $sif;
        my $buf;
        if (exists $s->{$chunk}) {
            $buf = $s->{$chunk};
        } elsif ($sif = $source[_n_sif_id($e)]) {
            $buf = $sif->_read($left > OutSz ? OutSz : $left, _n_offset($e));
        } else {
            die "No buf for chunk $chunk";
        }
        my $nr = _n_refs_count($e);
        my $o = _n_base_size;
        while ($nr--) {
            my ($off, $ref, $step)
                = unpack('Na6n', substr($e, $o, _n_ref_size));
            my $link = pack('V', $s->{remap}->{$ref});
            for my $i (0 .. 2) {
                substr($buf, $off, 4, $link);
                $off += $step;
            }
            $o += _n_ref_size;
        }
        # Set LogicalChannel if it's a StandardArea
        substr($buf, 48, 4, pack('V', _n_logicalchannel($e)))
            unless _n_is_data($e);
        while ($left > 0) {
            print $fh $buf or return undef;
            $cout += length $buf;
            $left -= length $buf;
            $buf = $sif->_read($left > OutSz ? OutSz : $left) if $left > 0;
        }
    }
    return ($cout == $s->{off});
}

sub sif_size {
    my ($s) = @_;
    return $s->{off};
}

# The following is a memory-saving device for the values of the
# $s->{map} hash. Those values used to be a hash of simple fields,
# with an additional child hash for the offset table. The code below
# allows all that to be encoded into a single string, using much
# less memory than the hash-hash implementation, at the cost of code
# clarity, complexity, and perhaps execution time.

sub _n_new { "\x00" x _n_base_size }

sub _n_logicalchannel { unpack 'V', $_[0] }
sub _n_set_logicalchannel { substr $_[0], 0, 4, pack('V', $_[1]) }

sub _n_offset { unpack 'x4V', $_[0] }
sub _n_set_offset { substr $_[0], 4, 4, pack('V', $_[1]) }

sub _n_size { unpack 'x8V', $_[0] }
sub _n_set_size { substr $_[0], 8, 4, pack('V', $_[1]) }

sub _n_sif_id { unpack 'x12v', $_[0] }
sub _n_set_sif_id { substr $_[0], 12, 2, pack('v', $_[1]) }

sub _n_is_data { unpack 'x14C', $_[0] }
sub _n_set_is_data { substr $_[0], 14, 1, pack('C', $_[1]) }

sub _n_remapped { unpack 'x15C', $_[0] }
sub _n_set_remapped { substr $_[0], 15, 1, pack('C', $_[1]) }

sub _n_add_3ref {
    my (undef, $id, $offset, $off, $step) = @_;

    my $node = $_[0];
    my $nr = (length($node) - _n_base_size) / _n_ref_size;

    my $newref = pack('NnNn', $off, $id, $offset, $step);
    my $o = _n_base_size;
    my $done = 0;
    if ($nr) {
        my $poff = pack('N', $off);
        while ($nr--) {
            if (substr($node, $o, 4) eq $poff) {
                substr($_[0], $o, _n_ref_size, $newref);
                $done = 1;
                last;
            }
            $o += _n_ref_size;
        }
    }
    $_[0] .= $newref unless $done;
}

sub _n_put_sa_ref {
    my ($s, undef, $prev, $self, $next) = @_;

    my ($id, $off) = unpack(DecomposeChunk, $self);
    _n_add_3ref($_[1], $id, $off, 0, 16); # SAP.self, SAP2.self, SAP3.self
    ($id, $off) = unpack(DecomposeChunk, $next);
    _n_add_3ref($_[1], $id, $off, 8, 16); # SAP.next, SAP2.next, SAP3.next
    ($id, $off) = unpack(DecomposeChunk, $prev);
    _n_add_3ref($_[1], $id, $off, 12, 16); # SAP.prev, SAP2.prev, SAP3.prev
}

sub _n_refs_values {
    my $nr = (length($_[0]) - _n_base_size) / _n_ref_size;
    my $o = _n_base_size + 4;
    my @v;
    while ($nr--) {
        push @v, substr($_[0], $o, 6);
        $o += _n_ref_size;
    }
    return sort @v;
}

sub _n_refs_count {
    return (length($_[0]) - _n_base_size) / _n_ref_size;
}

# node refs map  node_offset => [sif, offset]
# $s->{remap} maps [sif, offset] => output_offset

1;

