use strict;
use warnings 'all';

package SoMat::TCE::File;

sub new {
    my ($class) = @_;
    my $s = {};
    bless $s, ref $class || $class;
    $s->clear;
    return $s;
}

sub clear {
    my ($s) = @_;
    $s->{section_orders} = [];
    $s->{orders} = {};
    $s->{keys} = {};
}

sub add_section {
    my ($s, $section) = @_;
    unless (exists $s->{keys}{$section}) {
        push(@{$s->{section_orders}}, $section);
        $s->{orders}{$section} = [];
        $s->{keys}{$section} = {};
    }
}

sub sections {
    my ($s) = @_;
    return @{$s->{section_orders}};
};

sub keys { #KLUDGE - This should really be get_keys()
    my ($s, $section) = @_;
    return @{$s->{orders}{$section}};
};

sub get {
    my ($s, $sec, $key) = @_;
    if (defined $sec && defined $key) {
        return $s->{keys}{$sec}{$key};
    } elsif (defined $sec) {
        return $s->{keys}{$sec};
    } else {
        return $s->{keys};
    }
}

sub set {
    my ($s, $section, $key, $value) = @_;
    $s->add_section($section);
    unless (exists $s->{keys}{$section}{$key}) {
        push(@{$s->{orders}{$section}}, $key);
    }
    $s->{keys}{$section}{$key} = $value;
}

sub remove {
    my ($s, $section, $key) = @_;
    die "unimplemented";
}

sub remove_section {
    my ($s, $section) = @_;
    delete $s->{orders}{$section};
    delete $s->{keys}{$section};
    my @new = grep { $_ ne $section } @{$s->{section_orders}};
    $s->{section_orders} = \@new;
}

sub write {
    my ($s, $fh, $nl) = @_;
    $fh = *STDOUT unless defined $fh;
    $nl = "\n" unless defined $nl;
    my $sec_no = 0;
    for my $section (@{$s->{section_orders}}) {
        print $fh "$nl" if $sec_no++;
        print $fh "[$section]$nl";
        for my $key (@{$s->{orders}{$section}}) {
            print $fh "$key=$s->{keys}{$section}{$key}$nl";
        }
    }
    return $s;
}

sub write_file {
    my ($s, $name) = @_;
    my $fh;
    open($fh, '>', $name) || die "couldn't write $name: $!";
    $s->write($fh);
    close($fh);
    return $s;
}

sub write_string {
    my ($s, $str) = @_;
    my $fh;
    open($fh, '>', \$str) || die "couldn't write string: $!";
    $s->write($fh);
    close($fh);
    return $str;
}

sub read {
    my ($s, $fh) = @_;
    my $n = 0;
    my $section;
    while (<$fh>) {
        my $line = $_;
        chomp $line;
        $line =~ s/\r$//;
        ++$n;
        if ($line =~ /^\[(\w+)\]\s*$/) {
            $section = $1;
            next;
        }
        if ($line =~ /^([\w\.]+)=(.*)$/) {
            unless (defined $section) {
                die("Value not in section on line $n : $line\n");
            } else {
                $s->set($section, $1, $2);
            }
            next;
        }
        next if ($line =~ /^\s*$/);
        next if ($line =~ /^\s*\;.*$/);
        die("Syntax error on line $n : $line\n");
    }
    return $s;
}

sub read_file {
    my ($s, $name) = @_;
    defined $name or die "read_file() no name argument provided.";
    my $fh;
    open($fh, '<', $name) || die "couldn't read $name: $!";
    $s->read($fh);
    close($fh);
    return $s;
}

sub read_string {
    my ($s, $str) = @_;
    my $fh;
    open($fh, '<', \$str) || die "couldn't read string: $!";
    $s->read($fh);
    close($fh);
    return $s;
}

sub read_current {
    my ($s) = @_;
    require SoMat::Network::WebRequest;
    my $url = 'http://127.0.0.1/-/test/_DEFAULT_/get.tce';
    my $web = SoMat::Network::WebRequest->new($url);
    my $body = "";
    if ($web->get_response_code == 200) {
        $body = $web->getall;
    }
    $s->read_string($body);
}
 
sub mangle {
    my ($s) = @_;
    for my $sec ($s->sections) {
        for my $key ($s->keys($sec)) {
            local $_ = $s->get($sec, $key);
            s/^ ([\+-]?\d)/$1/;
            s/([\+-]?)(\d*\.?\d+)e([\+-])(\d\d(:?\D|$))/$1${2}e${3}0$4/;
            $s->set($sec, $key, $_);
        }
    }
    return $s;
}

sub remap_aux_files {
    my ($s, $auxdir) = @_;

    my $cs;
    for my $sec ($s->sections) {
        for my $key ($s->keys($sec)) {
            next unless $key =~ /^File(Name|Name2|Path)$/;

            require SoMat::eDAQ::ContentStore;
            $cs ||= SoMat::eDAQ::ContentStore->new(dir=>$auxdir);
            my $winpath = $s->get($sec, $key);
            my $edaqpath = $cs->get_content_filename(name=>$winpath);
            if (defined $edaqpath) {
                $s->set($sec, $key, $edaqpath);
            }
        }
    }
}

my @numbered_secs =
    qw(NetNode HardItem SmartItem ChanItem SoftItem DataItem);

sub merge {
    my ($s, @rest) = @_;
    my $last_sec = {};
    # The goal of this last_sec crap is so if the merge looks like
    # it's happening into a 'real' TCE file with a [Main] section,
    # to update the NumWhatsits keys.
    if (exists $s->get->{Main}) {
        for my $sec (@numbered_secs) {
            $last_sec->{$sec} = $s->get('Main', "Num${sec}s");
        }
    }
    for my $in (@rest) {
        for my $sec ($in->sections) {
            my $sec_class;
            $sec_class = $1 if $sec =~ /(\w+)_\d+/;
            my $dest_num = ++$last_sec->{$sec_class};
            my $dest_sec = $sec;
            $dest_sec = "${sec_class}_$dest_num" if $sec_class;
            for my $key ($in->keys($sec)) {
                $s->set($dest_sec, $key, $in->get($sec, $key));
            }
        }
    }
    if (exists $s->get->{Main}) {
        for my $sec (@numbered_secs) {
            $s->set('Main', "Num${sec}s", $last_sec->{$sec});
        }
    }
    return $s;
}
=pod  I don't think I need this... use length of get_nodes instead
sub is_network_file {
    my ($s) = @_;
    my $nodeCount = 0;
    for my $sec ($s->sections) {
        if ($sec =~ /^NetNode_\d/) {
           $nodeCount++;
        }
    }
    $nodeCount > 1 ? return $nodeCount: return 0;
}
=cut
 
sub get_network_nodes {
    my ($s) = @_;
    my %nodes;
    for my $sec ($s->sections) {
        if (exists $s->{keys}{$sec}{NodeName}) {
            $nodes{$s->{keys}{$sec}{NodeName}} = 1;
        }
    }
    return CORE::keys %nodes; #KLUDGE - part of the keys() vs get_keys() KLUDGE
}

sub get_slaves {
    my ($s) = @_;
    my $slaves = [];
    for my $sec ($s->sections) {
        if ( exists $s->{keys}{$sec}{NodeType} && $s->{keys}{$sec}{NodeType} == 2) {
            push (@{$slaves}, $s->{keys}{$sec}{NodeName});
        }
    }
    return $slaves;
} 

sub get_nodename_from_serial {
    my ($s, $serial) = @_;
    $serial or die "get_nodename_from_serial('MPB serial num') no serial number provided.";
    for my $sec ($s->sections) {
        if (    exists $s->{keys}{$sec}{Prefix} &&  
                      ('MS_MPB' eq $s->{keys}{$sec}{Prefix}) && 
                exists $s->{keys}{$sec}{SN} && 
                      ($serial eq $s->{keys}{$sec}{SN})) {
            return $s->{keys}{$sec}{NodeName};
        }
    }
    return undef; 
}

sub get_serial_from_nodename {
    my ($s, $nodename) = @_;
    $nodename or die "get_serial_from_nodename('NodeName') no nodename provided.";
    
    for my $sec ($s->sections) {
        if (    exists $s->{keys}{$sec}{Prefix} &&
                      ('MS_MPB' eq $s->{keys}{$sec}{Prefix}) && 
                exists $s->{keys}{$sec}{NodeName} && 
                      ($nodename eq $s->{keys}{$sec}{NodeName})) {
            return $s->{keys}{$sec}{SN};
        }
    }
    return undef;
}

sub get_nodetype_from_serial {
    my ($s, $serial) = @_;
    $serial or die "get_nodetype_from_serial('MPB serial num') no serial number provided.";
    my $nodename = $s->get_nodename_from_serial($serial);
 
    for my $sec( grep (/NetNode/, $s->sections)) {
        if ($s->{keys}{$sec}{NodeName} eq $nodename) {
            return $s->{keys}{$sec}{NodeType};
        }
    }
    return undef; 
}

sub make_my_tce_file {
    my ($s, $serial_number) = @_;
    $serial_number or die "make_my_tce_file('MPB Serial Number'): No serial number provided.";
    my $nodename = $s->get_nodename_from_serial($serial_number);
    for my $sec ($s->sections) {
        if (defined $s->{keys}{$sec}{NodeName} && $s->{keys}{$sec}{NodeName} ne $nodename) {
            $s->remove_section($sec);
        } 
    }
    #Renumber numbered sections and correct the counts in [Main]
    # Copies of global variable for rebuilding data structure.
    my $new_section_orders = [];
    my $new_orders = {};
    my $new_keys = {};

    my $section_count = {};
    my $index_count = 0;

    foreach my $section (@{$s->{section_orders}}) {
        unless ( $section =~ /^\w+_\d+$/ ) { # Numbered Section
            $new_section_orders->[$index_count] = $section;
            $new_keys->{$section} = $s->{keys}{$section};
            $new_orders->{$section} = $s->{orders}{$section};

            $index_count++;
        } else { 
            my $sec_rename = $section;
            $sec_rename =~ s/^(\w+?)_\d+$/"$1_" . ++$section_count->{$1}/e;

            $new_section_orders->[$index_count] = $sec_rename;
            $new_keys->{$sec_rename} = $s->{keys}{$section};
            $new_orders->{$sec_rename} = $s->{orders}{$section};

            $index_count++;
        }
    } 

    $s->{section_orders} = $new_section_orders;
    $s->{keys} = $new_keys;
    $s->{orders} = $new_orders;
    for my $sec (@numbered_secs) {
        $s->set('Main', "Num${sec}s", $section_count->{$sec} || '0');
    }
    return 0;
}

sub set_all_if_exists {
    my ($s, $key, $name) = @_;
    for my $sec ($s->sections) {
        if (defined $s->get($sec, $key)) {
            $s->set($sec, $key, $name);
        }
    }
}

1;

=head
{
    my $f = SoMat::TCE::File->new;
    $f->set('Main', 'TCEVersion', 'V3.8.7-RC2');
    $f->set('Main', 'TargetFCS', '1.2');
    $f->set('NetNode_1', 'NodeName', 'edaq16');
    $f->write(*STDOUT);
    $f->clear;
    open(F, $ARGV[0]);
    $f->read(*F);
    close(F);
    $f->write(*STDOUT);
}
=cut
