
package SoMat::eDAQ::ContentStore;

=head1 NAME

SoMat::eDAQ::ContentStore - Simple content-based storage

=head1 SYNOPSIS

    use SoMat::eDAQ::ContentStore;

    my $cs_dir = '/hd/content_store';
    my $cs = SoMat::eDAQ::ContentStore->new(dir=>$cs_dir);

    my $hash1 = cs->put(name=>'c:\foo\bar', filename=>'/tmp/foobar');
    my $hash2 = cs->put(name=>'c:\foo\blort', content=>"File content...\n");

    my $list = $cs->list;
    for my $pair (@$list) {
        my ($name, $md5hash) = @$pair;

        my $filename = $cs->get_content_filename($name);
        print "name=$name, md5hash=$md5hash, filename=$filename\n";

        my $fh = $cs->get_content_fh($name);
        # do something with $fh here
        close($fh);
    }

    $cs->clear; # Remove everything. Be Very Sure you want to do this.
    # Or alternatively...
    SoMat::eDAQ::ContentStore->new(dir=>$cs_dir, clear=>1);

=cut

use strict;
use warnings 'all';
require DBI;
require SoMat::File;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    die "Must define 'dir' argument" unless defined $s->{dir};
    $s->{dbfile} = "$s->{dir}/index.sqlite";
    $s->{table} ||= 'content_map';
    $s->clear if $s->{clear};
    $s->init();
    return $s;
}

sub dbh {
    my ($s) = @_;
    return $s->{dbh} if defined $s->{dbh};
    $s->{dbh} = DBI->connect("dbi:SQLite:dbname=$s->{dbfile}", "","",
        {RaiseError=>1, AutoCommit=>0});
    return $s->{dbh};
}

sub init {
    my ($s) = @_;

    unless (-d $s->{dir}) {
        mkdir($s->{dir}, 0777) or die "Can't mkdir $s->{dir}: $!";
    }

    # KLUDGE - DBD::SQLite quote_identifier() broken
    $s->{qtable} = $s->{table};

    # KLDUGE we can do this better with updated SQLITE (IF NOT EXISTS)
    unless (-f $s->{dbfile}) {
        my $dbh = $s->dbh;
        #$s->{qtable} = $dbh->quote_identifier($s->{table});
        # KLUDGE - DBD::SQLite quote_identifier() broken
        #$s->{qtable} = $s->dbh->quote_identifier($s->{qtable});
        $s->{qtable} = $s->{table};
        $dbh->do(qq{
            CREATE TABLE $s->{qtable} (
                name text,
                hash text
            );
        });
        $dbh->commit;
    }
    # KLUDGE - DBD::SQLite quote_identifier() broken
    #$s->{qtable} = $s->dbh->quote_identifier($s->{qtable});
    $s->{qtable} = $s->{table};
}

sub put {
    my ($s, %args) = @_;

    my $tmpname = $args{filename};
    unless (defined $tmpname) {
        $tmpname = SoMat::File::Temp->tempfile
            or die "Can't make temp file: $!";
        open(my $fh, '>', $tmpname) or die "Can't write temp file: $!";
        print $fh $args{content};
        close($fh) or die "Can't close temp file: $!";
    }

    open(my $hashfh, '-|', '/usr/bin/md5sum', $tmpname)
        or die "Can't md5sum";
    my $hash = <$hashfh>;
    close($hashfh);
    $hash =~ s/^([0-9a-f]+).*$/$1/s;
    die "md5sum failed" unless $hash;

    unless (-f "$s->{dir}/$hash") {
        system('/bin/cp', $tmpname, "$s->{dir}/$hash")
            and die "Can't cp to $s->{dir}/$hash: $!";
    }
    unlink($tmpname) unless $args{filename};

    my $dbh = $s->dbh;

    my ($qname, $qhash) = ($dbh->quote($args{name}), $dbh->quote($hash));

    my $sql;
    # KLUDGE this has to be done as two separate do() requests to work,
    # probably due to buggage in DBD::SQLite
    $sql = qq{
        DELETE FROM $s->{qtable} WHERE name = $qname;
    };
    $dbh->do($sql);
    $sql = qq{
        INSERT INTO $s->{qtable} (name, hash) VALUES ($qname, $qhash);
    };
    $dbh->do($sql);
    $dbh->commit;
    return $hash;
}

sub get_hash_from_name {
    my ($s, $name) = @_;

    my $dbh = $s->dbh;
    my $qname = $dbh->quote($name);
    my $sql = "SELECT hash FROM $s->{qtable} WHERE name = $qname";
    my $col = $dbh->selectcol_arrayref($sql);
    $dbh->commit;
    return undef unless ref($col) eq 'ARRAY';
    return $col->[0];
}

sub get_content_filename {
    my ($s, %args) = @_;

    my $hash = $s->get_hash_from_name($args{name});
    return undef unless defined $hash;
    return "$s->{dir}/$hash";
}

sub get_content_fh {
    my ($s, %args) = @_;

    open(my $fh, '<', $s->get_content_filename($args{name}));
    return $fh;
}

sub list {
    my ($s) = @_;

    my $dbh = $s->dbh;
    my $sql = "SELECT name, hash FROM $s->{qtable} ORDER BY name;";
    my $a = $dbh->selectall_arrayref($sql);
    $dbh->commit;
    return $a;
}

sub clear {
    my ($s) = @_;

    if (-e $s->{dir}) {
        system('/bin/rm', '-rf', $s->{dir})
            and die "rm -rf $s->{dir} failed: $!";
    }
    if ($s->{dbh}) {
        $s->{dbh}->disconnect;
        delete $s->{dbh};
    }
}

1;
