package SoMat::eDAQ::Hardware::MultiUpdate;

require SoMat::File;
require SoMat::eDAQ::Hardware::Update;
require SoMat::eDAQ::Hardware::Expmod;
require SoMat::eDAQ::Hardware::Update::AVT;
require SoMat::eDAQ::Hardware::VBM;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    return $s;
}

sub open_maybe_gzipped {
    my ($s, $bundle, $filename) = @_;

    my $cmd = "/usr/bin/unzip -p '$bundle' '$filename'";

    my $fh = SoMat::File->new("$cmd|")
        or die("Can't read $filename from $bundle: $!");
    my $gzipped;
    $fh->read($gzipped, 2) == 2 or die("Can't read $filename from $bundle: $!");
    $gzipped = ($gzipped eq "\x1F\x8B") ? 1 : 0;
    undef $fh;

    open($fh, ($gzipped ? "$cmd | /bin/gunzip -c |" : "$cmd|"))
        or die("Can't read $filename from $bundle: $!");
    return $fh;
}

sub image_size {
    my ($s, $bundle, $filename) = @_;
    my $bytes = 0;
    my $lines = 0;
    if (my $fh = $s->open_maybe_gzipped($bundle, $filename)) {
        while (defined (my $line = <$fh>)) {
            ++$lines;
            $bytes += length($line);
        }
        close($fh);
    }
    return ($bytes, $lines);
}

sub copy_file_from_bundle {
    my ($s, $bundle, $filename, $dest) = @_;
    system("/usr/bin/unzip -p '$bundle' '$filename' > '$dest'") == 0
        or die "Could not copy system firmware image to /";
}

sub update_exp {
    my ($s, $hw, $bundle, $filename) = @_;
    my $dev = $hw->{device};

    my $upper = SoMat::eDAQ::Hardware::Update->new(%$hw);
    my ($image_bytes, $image_lines) = $s->image_size($bundle, $filename);
    $upper->init_progress($image_lines);
    my $fh = $s->open_maybe_gzipped($bundle, $filename);

    my $devicename = $hw->{device};
    my $port = 0xff0d;

    my $device = SoMat::eDAQ::Hardware::Expmod->new($devicename);
    die "Failed to open $devicename: $!\n" unless $device;
    $device->set_timeout(5.0); # 5 seconds
    $device->set_port($port);
    my $scriptline = 0;

    my $false_if = 0;

    my $progress_message = '';
    while (<$fh>) {
        $scriptline++;
        $upper->progress_update($scriptline);
        $upper->progress($progress_message);
        s/\r?\n?$//o;
        if (substr($_, 0, 1) eq '.') {
            if (/^\.IF\s+(\S+)\s*=\~\s*\/(\S+)\/\s*/oi) {
                my ($var, $val) = ($1, $2);
                my $resp = $device->get_named_variable($var);
                print "var $var val $val resp $resp\n";
                if ($resp !~ /$val/) {
                    $false_if = 1;
                }
            } elsif (/^\.IF\s+(\S+)\s*=\s*(\S+)\s*/oi) {
                my ($var, $val) = ($1, $2);
                my $resp = $device->get_named_variable($var);
                if ($resp ne $val) {
                    $false_if = 1;
                }
            } elsif (/^\.ELSE/oi) {
                $false_if = !$false_if;
            } elsif (/^\.ENDIF/oi) {
                $false_if = 0;
            } elsif (!$false_if && /^\.MESSAGE\s+(.*)$/oi) {
                #print "$1\n";
                $progress_message = $1;
                $upper->progress($1);
            } elsif (!$false_if && /^\.DELAY\s+(.*)$/oi) {
                sleep $1;
            }
            next;
        }
        my $command = $_;
        next unless $command;
        next if $false_if;
        $device->send_packet($command);
        my $response;
        my $timeout = 10;
        while ($timeout && !($response = $device->recv_cmd_response())) {
            print "$hw->{serial}: Timeout in script line $scriptline ",
                "($command), trying again...\n";
            $timeout--;
            $device->send_packet($command);
        }
        die "Error in script line $scriptline ($command): Timed out\n"
            unless $response;
        for my $line (split("\n", $response)) {
            if ($line =~ /^ERRNO\=(.*)/) {
                die "$hw->{serial}: Error in script line $scriptline ($command): $line\n"
                    if $1 ne 'OK';
            } else {
                $upper->progress($line) if $line;
            }
        }
    }
    $upper->progress("Finished");

    return 1;
}

sub update_hlb {
    my ($s, $hw, $bundle, $filename) = @_;

    eval {
        my ($image_bytes, $image_lines) = $s->image_size($bundle, $filename);
        my $fh = $s->open_maybe_gzipped($bundle, $filename);

        my $avt = SoMat::eDAQ::Hardware::Update::AVT->new(
            tcename=>$hw->{tcename}, device_name=>$hw->{device}, debug=>0,
            no_device=>0, test_mode=>0, erase=>1, do_write=>1, verify=>0);

        eval { $avt->reflash($fh, $image_bytes); };
        #$avt->reflash($fh, $image_bytes);
        if ($@) {
            $avt->progress("$hw->{tcename} ERROR: $@");
        }
    };

    if ($@) { print "$hw->{tcename} ERROR: $@\n"; }
    #if ($@) { $avt->progress("$hw->{tcename} ERROR: $@"); }
}

sub update_mpb {
    my ($s, $hw, $bundle, $filename) = @_;

    my $upper = SoMat::eDAQ::Hardware::Update->new(%$hw);
    $upper->init_progress(2);
    $upper->progress_update(0);
    $upper->progress("Verifying firmware image");

    my $dest = '/__e10-update.tar.gz'; # KLUDGE hardcode

    eval {
        $s->copy_file_from_bundle($bundle, $filename, $dest);
        $upper->progress_update(1);
        $upper->progress("Copying firmware image to FLASH");

        my $output = `gunzip -t '$dest' 2>&1`;
        die "System firmware image verification failed: $output" if $?;
    };
    if ($@) {
        unlink $dest;
        open my $f, '>', '/eDAQ/ramdisk.size';
        print $f int((-s '/eDAQ/ramdisk') / 512);
        close $f;
        unlink '/eDAQ/ramdisk';

        $s->copy_file_from_bundle($bundle, $filename, $dest);
        $upper->progress_update(1);
        $upper->progress("Copying firmware image to FLASH");

        my $output = `gunzip -t '$dest' 2>&1`;
        die "System firmware image verification failed: $output" if $?;
    }

    $upper->progress_update(2);
    $upper->progress("Firmware image copied to FLASH; update occurs on reset");

    return 1;
}

sub update_pwr {
    my ($s, $hw, $bundle, $filename) = @_;

    my $upper = SoMat::eDAQ::Hardware::Update->new;
    $upper->init_progress(100);
    $upper->progress('');

    #cleanup;
    unlink "/tmp/power_firmware.hex.gz";
    unlink "/tmp/power_firmware.hex";

    my $out;
    my $hexfile = '/tmp/power_firmware.hex';
    my $hexgz = "$hexfile.gz";
    $s->copy_file_from_bundle($bundle, $filename, $hexgz);
    $out = `gunzip -c $hexgz | grep '^S01' 2>&1`;
    die "Power Controller firmware image verification failed: $out\n" if $?;
    $out = `gunzip $hexgz 2>&1`;

    die "Failed to make firmware hexfile: $out\n" if $?;
    my $powerd_pid = `cat /var/run/powerd.pid`;
    $out = `kill $powerd_pid 2>&1`;
    die "Failed to kill power daemon: $out\n" if $?;
    sleep 1;

    $upper->progress_update(0);
    $upper->progress("Writing FLASH");
    open(my $pwrd, "/sbin/powerd -spammy-load $hexfile|");
    while (defined (my $line = <$pwrd>)) {
        $line =~ s/^\s*(.*?)\s*$/$1/s;
        if ($line =~ /^(\d+)(\.\d+)?\% complete/) {
            $upper->progress_update($1 * 0.9);
            $upper->progress("Writing FLASH");
        } else {
            print "$line didn't match regexp...\n";
        }
    }
    close $pwrd;

    sleep 5;
    my $pid = fork;
    die "Fork failed: $@" unless defined $pid;
    if ($pid == 0) {
        syscall(66); # SYS_setsid
        for my $fd (0 .. 20) {
            syscall(6, $fd); # SYS_close - KLUDGE!!!
        }
        exec "/sbin/powerd";
    }
    sleep 1;
    my $state = `pwrcmd 'APPSTATE?'`;
    chomp $state;
    system "kill -USR1 `cat /var/run/powerd.pid`"; # reset LEDs
    if ($state eq '0|RUN') {
        $upper->progress_update(100);
        $upper->progress("Power controller firmware update successful");
    } else {
        die "Power controller firmware update NOT successful."
            . " Contact SoMat for further instructions!";
    }
    return 1;
}

sub update_reset {
    my ($s, $hw, $bundle, $filename) = @_;

    my $upper = SoMat::eDAQ::Hardware::Update->new(%$hw);
    $upper->init_progress(100);
    $upper->progress_update(0);
    $upper->progress("");

    my $msg = <<"EOP";
<p>
The eDAQ is being reset.
</p>
<ul>
<li> The eDAQ will be rebooted.  After the reboot all three LEDs will
     be on.</li>
<li> As the eDAQ is rebooting, the system firmware update will take place.
     The red LED will be blinking at 2 Hz while this is happening.  
     <b>Do NOT remove power to the eDAQ or toggle the
     front panel or remote power switch during this time!</b>
     This may take several minutes.  (Diagnostic output will appear on the 
     serial port at 115200 bps.)</li>
<li> After the update is complete, the eDAQ will boot normally.</li>
</ul>
<p>
When the eDAQ has successfully rebooted, click the "OK" button to
continue.
</p>
EOP

    # KLUDGE maybe we should try to play some fun Javascript games
    # with $.ajax() to tell when the eDAQ has really reset. Won't
    # work unless we know for sure what address the client is
    # actually using to talk to the eDAQ.

    $upper->progress_update(1);
    $upper->progress("Resetting eDAQ");
    $upper->message(message=>$msg);

    my $powerd_pid = `cat /var/run/powerd.pid`;
    my $output = `kill -PWR $powerd_pid 2>&1`;
    die "Failed to reboot system: $output" if $?;

    return 1;
}

sub update_vbm {
    my ($s, $hw, $bundle, $filename) = @_;
    my $dev = $hw->{device};

    my $upper = SoMat::eDAQ::Hardware::Update->new;
    my ($image_bytes, $image_lines) = $s->image_size($bundle, $filename);
    $upper->init_progress($image_lines);
    my $fh = $s->open_maybe_gzipped($bundle, $filename);
    $upper->progress('Initializing');

    my $dio = SoMat::eDAQ::Hardware::Expmod->new($dev);

    my $address = $hw->{module_address};
    my ($bank) = $address =~ /LUN=IO_BANK_(\w+)/i;
    #print "ADDRESS = $address\n";
    $bank = '0' if $address =~ /LUN=CAN/i;
    $bank = '1' if $address =~ /LUN=VBM/i;
    my ($channel) = $address =~ /CHANNEL=(\d+)/i;
    # KLUDGE should use current VBM_CONTROL_PORT if already set
    my $port = eval "0x24$bank$channel";
    #printf "PORT = $port = 0x%X\n", $port;

    # More than one of these may happen at once, so we need a unique
    # expmod command port!
    my $eport = eval "0xff$bank$channel";
    $dio->set_port($eport);

    uc($dio->get_named_variable("VBM_PRESENT", $address)) eq 'YES'
        or die("No VBM present at address '$address'");

    $dio->send_packet("SET $address VBM_CONTROL_PORT=$port VBM_ENABLE=YES");
    my $resp = $dio->recv_cmd_response();
    $resp =~ /^ERRNO=OK$/i
        or die("Couldn't enable VBM module: $resp");

    my $vbm = SoMat::eDAQ::Hardware::VBM->new($dio, $port);

    $upper->progress("Updating VBM %s; previous firmware %s",
        $s->vbm_cmd($vbm, "SERNUM?"), $s->vbm_cmd($vbm, "VERSION?"));

    #$upper->progress("APPSTATE=%s", $s->vbm_cmd($vbm, "APPSTATE?"));

    $s->vbm_cmd($vbm, 'APPSTATE=LOADER');
    for my $i (-10 .. 0) {
        $i or die("Can't invoke LOADER");
        last if uc($s->vbm_cmd($vbm, "APPSTATE?")) eq 'LOADER';
        sleep 1;
    }

    $upper->progress("Erasing FLASH");
    $s->vbm_cmd($vbm, 'FLASH_ERASE!');

    $upper->progress("Writing FLASH");
    my $iter = 0;
    my $scriptline = 0;
    while (defined (my $line = <$fh>)) {
        $scriptline++;
        $upper->progress_update($scriptline);
        $upper->progress("Writing FLASH");

        next unless $line =~ /^S[^0]/;
        $line =~ s/[\x0D\x0A]$//;
        my $cmd = 'LINESTART';
        while (length($line)) {
            my $part = substr($line, 0, 20, '');
            $s->vbm_cmd($vbm, "$cmd=$part");
            $cmd = 'LINECONT';
        }
        $s->vbm_cmd($vbm, 'FLASH_WRITE!');
    }
    $upper->progress("Reinitalizing module");

    $s->vbm_cmd($vbm, 'APPSTATE=RUN');

    for my $i (-10 .. 0) {
        $i or die("Can't exit LOADER");
        my $st = $s->vbm_cmd($vbm, "APPSTATE?");
        defined $st or die("APPSTATE? failed");
        last if $st && $st ne 'LOADER';
        sleep 1;
    }

    my $msg = sprintf "Successfully updated firmware for VBM %s to %s (%s)",
        $s->vbm_cmd($vbm, "SERNUM?"), $s->vbm_cmd($vbm, "VERSION?"),
        $s->vbm_cmd($vbm, "APPSTATE?");
    $upper->progress($msg);
}

sub vbm_cmd {
    my ($s, $vbm, $cmd) = @_;
    return $vbm->run_command_die($cmd);
}

sub get_manifest {
    my ($s, $bundle_file) = @_;

    my $mf = eval `/usr/bin/unzip -p '$bundle_file' '*/updater_manifest.pl'`;
    if ($@) {
        die "Can't get manifest file";
    }
    return $mf;
}

1;
