#!/usr/bin/perl

use strict;

use SoMat::eDAQ::Hardware::Expmod;
use SoMat::eDAQ::Hardware::Eeprom;

package SoMat::eDAQ::Hardware::Query::Hwdb;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $s  = {};
    $s->{db} = [ ];
    bless ($s, $class);
    return $s;
}

sub add {
    my $s = shift;
    my %keys = @_;

    push(@{$s->{db}}, \%keys);
    return \%keys;
}

sub get {
    my $s = shift;
    return $s->{db};
}

sub next_uid {
    my $s = shift;
    $s->{next_uid} ||= 0;
    return $s->{next_uid}++;
}

sub get_columns {
    my $s = shift;
    my %cols;

    my @first = qw(tcename device serial version build ecns);
    my @rest;

    foreach my $row (@{$s->{db}}) {
        foreach my $col (keys %$row) {
            $cols{$col}++;
        }
    }

    foreach my $first_el (@first) {
        delete $cols{$first_el};
    }

    foreach my $col (keys %cols) {
        push(@rest, $col);
    }

    @first = (@first, sort @rest);

    return \@first;
}

sub get_tabdb {
    my $s = shift;
    my $cols = $s->get_columns();
    my $tabdb = join("\t", @$cols) . "\n";
    foreach my $row (@{$s->{db}}) {
        $tabdb .= join("\t", map { 
            if ($_ eq 'ecns') {
                join(",", map { $_->{number} . ";" . $_->{date} }
                     @{$row->{ecns}->{ecns}});
            } else {
                $row->{$_};
            }
        } @$cols) . "\n";
    }
    return $tabdb;
}

sub add_device {
    my ($s, $devref, $board_info, %other_keys) = @_;
    my $device = $devref->{device};
    my $devno = 0;
    $devno = $1 if ($device =~ /^.*?(\d+)$/);
    my $ret = { };

    if (exists $board_info->{do_if}) {
        return if !&{$board_info->{do_if}}($ret, $devref);
    }
    $ret->{device} = exists $board_info->{device}
        ? sprintf $board_info->{device}, $devno
        : $device;
    $ret->{serial} = $devref->get_serial();
    $ret->{description} = $board_info->{description};
    
    my $tcename = $board_info->{tcename};
    if ($devref->{type} eq "Expmod") {
        my $app_name = $devref->get_named_variable("APP_NAME");
        if ($app_name eq 'LOADER' || 
            $app_name eq 'PROGRAMMER') {
            $tcename = 'BadFW_%d';
        }
    }
    $ret->{tcename} = sprintf($tcename, $s->next_uid());

    $ret->{ecns} = $devref->get_ecns();
    if ($devref->{type} eq "Expmod") {
        $ret->{version} = $devref->get_version();
        $ret->{build} = $devref->get_build();
        $ret->{architecture} = 'expmod'; # Used to place the fw update
                                         # links on the web hw page
    } else {
        if (exists $board_info->{version}) {
            $ret->{version} = &{$board_info->{version}}($ret->{device});
        }
        if (exists $board_info->{build}) {
            $ret->{build} = &{$board_info->{build}}($ret->{device});
        }
    }
    if (exists $board_info->{slot}) {
        $ret->{address} = $devno;
    }
    $ret->{kernel_buffer_hint} = $board_info->{kernel_buffer_hint};
    $ret->{server_buffer_hint} = $board_info->{server_buffer_hint};

    if (exists $board_info->{extra}) {
        &{$board_info->{extra}}($ret, $devref);
    }

    my $line = $s->add(%$ret, %other_keys);

    if (exists $board_info->{after}) {
        &{$board_info->{after}}($s, $devref, $ret);
    }

    return $line;
}

sub renumber {
    my $s = shift;
    my $prefix = shift || "";
    my %last;
    my %changes;

    foreach my $line (@{$s->{db}}) {
        if ($line->{tcename} =~ /^(.*?)_([\d\.]+)$/) {
            $last{$1} ||= 1;
            my $new = $1 . "_" . $prefix . $last{$1}++;
            $changes{$line->{tcename}} = $new;
#print "Renumbering $line->{tcename} -> $new\n";
        }
    }

    foreach my $line (@{$s->{db}}) {
        foreach my $key (keys %$line) {
            CHANGE: foreach my $change (keys %changes) {
                if ($line->{$key} eq $change) {
#print "Updating old key $key in line $line->{tcename} from $change to new value $changes{$change}\n";
                    $line->{$key} = $changes{$change};
                    last CHANGE;
                }
            }
        }
    }
}

package SoMat::eDAQ::Hardware::Query;

sub make_scan_list {
    my @boards;

    my $FAKE_MPB10_DEVICE = 'fake-mpb10-device';
    my $FAKE_ELPWR_DEVICE = 'fake-elpwr-device';
    my $EEPROM_DEVICE = '/dev/edaq/eeprom/%d';
    my $EXPMOD_DEVICE = '/dev/edaq/expmod/%d';
    
    if ( -f '/eDAQ/.is-an-edaq-classic' ) {
        $EEPROM_DEVICE = '/dev/eeprom%d';
        $EXPMOD_DEVICE = '/dev/expmod%d';
    }

    unless (-f "/eDAQ/.is-an-edaq-classic") {
        push(@boards, [ $FAKE_MPB10_DEVICE, "SoMat::eDAQ::Hardware::eDAQ10MPB" ]);
    }

    if (-f "/eDAQ/.is-an-edaq-lite") {
        push(@boards, [ $FAKE_ELPWR_DEVICE, "SoMat::eDAQ::Hardware::eDAQLitePWR" ]);
    }
    
    for (my $i = 0; $i < 8; $i++) {
        push(@boards, [ sprintf($EEPROM_DEVICE, $i), "SoMat::eDAQ::Hardware::Eeprom" ]);
    }

    for (my $i = 0; $i < 8; $i++) {
        push(@boards, [ sprintf($EXPMOD_DEVICE, $i), "SoMat::eDAQ::Hardware::Expmod" ]);
    }

    return \@boards;
}

sub hw_query {
    my ($boards) = @_;
    $boards ||= '/etc/hardware.pm';
    $boards = do $boards unless ref($boards);

    my $scan_list = make_scan_list();
    my $db = SoMat::eDAQ::Hardware::Query::Hwdb->new();

    foreach my $scan (@$scan_list) {
        my ($device, $type) = @$scan;
        my $devref = $type->new($device);
        next unless $devref;
        my $serial =       # KLUDGE - this will retry on failure
            $devref->get_serial() || $devref->get_serial();
        my $board_type = "__UNKNOWN__";
        
#print "Got serial: $serial $devref->{type}\n";
        foreach my $key (keys %$boards) {
            if ($serial =~ /$key/) {
                $board_type = $key;
                last;
            }
        }
        my %extra;
        foreach my $board_info (@{$boards->{$board_type}}) {
            my $line = $db->add_device($devref, $board_info, %extra);
            if ($board_info->{slot} eq 'hardware') {
                $extra{parent} = $line->{tcename};
            }
        }
    }

    $db->renumber();
    return $db;
}

sub gen_hw_tab_db {
    my $db = hw_query(@_);
    return $db->get_tabdb();
}

package SoMat::eDAQ::Hardware::eDAQ10MPB;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};
    bless ($self, $class);
    return $self;
}

sub get_serial {
    my $serial = `e10-params -get factory serial-number`;
    chomp $serial;
    return $serial || "MSMPB.10-0000";
}

sub get_ecns {
    my $self = shift;
    my $ecn_line = `e10-params -ecn-list`;
    chomp $ecn_line;
    my %ret;
    $ret{count} = 0;
    foreach my $ecn_pair (split(",", $ecn_line)) {
        my ($ecn, $date) = split(";", $ecn_pair);
            $ret{ecns}->[$ret{count}]->{number} = $ecn;
            $ret{ecns}->[$ret{count}]->{date} = $date;
            $ret{count}++;
    }
    return \%ret;
}

package SoMat::eDAQ::Hardware::eDAQLitePWR;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};
    bless ($self, $class);
    return $self;
}

sub get_serial {
    my $serial = `pwrcmd 'SERIAL_NUMBER?'`;
    chomp $serial;
    $serial = $1 if ($serial =~ (/^0\|(ELPWR.*)$/));
    return $serial;
}

sub get_ecns {
    my $self = shift;
    my %ret;
    $ret{count} = 0;
    system "pwrcmd 'ECN!' >/dev/null";
    while ((my $ecn_pair = `pwrcmd 'ECN?'`) =~ /^0\|(\S+)$/) {
        $ecn_pair = $1;
        chomp $ecn_pair;
        my ($ecn, $date) = split("_", $ecn_pair);
        $ret{ecns}->[$ret{count}]->{number} = $ecn;
        $ret{ecns}->[$ret{count}]->{date} = $date;
        $ret{count}++;
    }
    system "pwrcmd 'ECN!' >/dev/null";
    return \%ret;
}

1;
