use strict;

package SoMat::eDAQ::Hardware::Serial;

require SoMat::File;

my $cc_chars = {
    'VINTR' => 0,
    'VQUIT' => 1,
    'VERASE' => 2,
    'VKILL' => 3,
    'VEOF' => 4,
    'VTIME' => 5,
    'VMIN' => 6,
    'VSWTC' => 7,
    'VSTART' => 8,
    'VSTOP' => 9,
    'VSUSP' => 10,
    'VEOL' => 11,
    'VREPRINT' => 12,
    'VDISCARD' => 13,
    'VWERASE' => 14,
    'VLNEXT' => 15,
    'VEOL2' => 16
};

my $iflags = {
    'IGNBRK' =>  0000001,
    'BRKINT' =>  0000002,
    'IGNPAR' =>  0000004,
    'PARMRK' =>  0000010,
    'INPCK' =>   0000020,
    'ISTRIP' =>  0000040,
    'INLCR' =>   0000100,
    'IGNCR' =>   0000200,
    'ICRNL' =>   0000400,
    'IUCLC' =>   0001000,
    'IXON' =>    0002000,
    'IXANY' =>   0004000,
    'IXOFF' =>   0010000,
    'IMAXBEL' => 0020000
};

my $oflags = {
    'OPOST' =>  0000001,
    'OLCUC' =>  0000002,
    'ONLCR' =>  0000004,
    'OCRNL' =>  0000010,
    'ONOCR' =>  0000020,
    'ONLRET' => 0000040,
    'OFILL' =>  0000100,
    'OFDEL' =>  0000200
};

my $cflags = {
    'CSIZE' =>  0000060,
    'CS5' =>    0000000,
    'CS6' =>    0000020,
    'CS7' =>    0000040,
    'CS8' =>    0000060,
    'CSTOPB' => 0000100,
    'CREAD' =>  0000200,
    'PARENB' => 0000400,
    'PARODD' => 0001000,
    'HUPCL' =>  0002000,
    'CLOCAL' => 0004000,
    'CRTSCTS'=> 020000000000
};

my $lflags = {
    'ISIG' =>   0000001,
    'ICANON' => 0000002,
    'ECHO' =>   0000010,
    'ECHOE' =>  0000020,
    'ECHOK' =>  0000040,
    'ECHONL' => 0000100,
    'NOFLSH' => 0000200,
    'TOSTOP' => 0000400,
    'IEXTEN' => 0100000
};

my $posix_speeds = {
    0       => 000,
    50      => 001,
    75      => 002,
    110     => 003,
    134     => 004,
    150     => 005,
    200     => 006,
    300     => 007,
    600     => 010,
    1200    => 011,
    1800    => 012,
    2400    => 013,
    4800    => 014,
    9600    => 015,
    19200   => 016,
    38400   => 017,
    # LINUX KLUDGE!!!!!!!
    57600   => 0010001,
    115200  => 0010002,
    230400  => 0010003,
    460800  => 0010004,
    500000  => 0010005,
    576000  => 0010006,
    921600  => 0010007,
    1000000 => 0010010,
    1152000 => 0010011,
    1500000 => 0010012,
    2000000 => 0010013,
    2500000 => 0010014,
    3000000 => 0010015,
    3500000 => 0010016,
    4000000 => 0010017
};

sub get_posix_speed {
    my ($speed) = @_;

    return exists($posix_speeds->{$speed}) ? $posix_speeds->{$speed} : 0;
}

sub stty {
    my $s = shift;
    my $set = shift; # KLUDGE check for ref($set) eq 'HASH'
    my @fds = @_;

    my @cc = 0 x 32;
    my ($iflag, $lflag, $cflag, $oflag, $speed) = 0 x 6;

    for my $setting (split(/\s+/, $set)) {
        my ($key, $val) = split(/=/, $setting);
        $val = 0 unless defined $val;
        $cc[$cc_chars->{$key}] = $val if exists $cc_chars->{$key};
        $iflag |= $iflags->{$key} if exists $iflags->{$key};
        $oflag |= $oflags->{$key} if exists $oflags->{$key};
        $cflag |= $cflags->{$key} if exists $cflags->{$key};
        $lflag |= $lflags->{$key} if exists $lflags->{$key};
        if ($key eq 'SPEED') {
            $speed = get_posix_speed($val);
            $cflag |= $speed;
        }
    }

    my $termios = pack("IIIICC32II", 
        $iflag, $oflag, $cflag, $lflag,
        0, @cc, $speed, $speed); # KLUDGE line discipline?

    for my $fd (@fds) {
        # FYI TCGETS = 0x5401
        ioctl($fd, 0x540B, 0); # TCFLSH
        die "TCSETS failed for fd $fd" if ioctl($fd, 0x5402, $termios) < 0;
    }
    return 1;
}

sub argv_to_stty {
    my $s = shift;
    my $stty = "";
    my $vals = {
        'baudrate' => 'SPEED=19200',
        'charactersize' => 'CS8',
        'parity' => 'IGNPAR',
        'stopbits' => '',
        'flowcontrol' => '',
        'modemcontrol' => 'CLOCAL'
    };

    # KLUDGE this is all wrong - should have defaults like 8 N 1 No Local
    while (@_) {
        local $_ = shift;
        $vals->{baudrate} = 'SPEED=' . shift(@_) if /^-baudrate$/;
        if (/^-charactersize$/) {
            my $cs = shift;
            $vals->{charactersize} = "CS$cs" if $cs =~ /^[5678]$/;
        }
        if (/^-parity$/) {
            my $parity = shift;
            $vals->{parity} = 'IGNPAR';
            $vals->{parity} = 'PARENB' if $parity =~ /^[EO]/i;
            $vals->{parity} = 'PARODD' if $parity =~ /^O/i;
        }
        if (/^-stopbits$/) {
            $stty .= 'CSTOPB' if shift(@_) == 2;
        }
        if (/^-flowcontrol$/) {
            my $flow = shift;
            $vals->{flowcontrol} = 'CRTSCTS' if $flow =~ /^H/i;
            $vals->{flowcontrol} = 'IXON IXANY IXOFF' if $flow =~ /^X/i;
        }
        if (/^-modemcontrol$/) {
            $vals->{flowcontrol} = 'CLOCAL' unless shift(@_) =~ /^1$/;
        }
    }
    return join(' ', 'IGNBRK CREAD HUPCL VMIN=1', values %$vals );
}

sub open {
    my $s = shift;
    my $path = shift;
    my $fh = SoMat::File->new($path, 04002); # O_NONBLOCK | O_RDWR;
#print STDERR "opened $path, fh = $fh\n";
    return undef unless defined $fh;
    fcntl($fh, 4, 02); # KLUDGE F_SETFL
    fcntl($fh, 2, 00); # KLUDGE F_SETFD no-close-on-exec
    my $stty = $s->argv_to_stty(@_);
#print STDERR "got stty string $stty\n";
    $s->stty($stty, $fh);
    return $fh;
}

1;
