package SoMat::eDAQ::Hardware::Update::AVT;

require SoMat::eDAQ::Hardware::Update;

our @ISA = qw(SoMat::eDAQ::Hardware::Update);

sub new {
    my $class = shift;
    my $s = $class->SUPER::new(@_);
    open($s->{dev}, '+<', $s->{device_name})
        or die "Can't open device $s->{device_name}: $!";
    return $s;
}

sub reset {
    my ($s) = @_;
    my $buf = '';
    $s->debug("Resetting $s->{dev}...\n");
    ioctl($s->{dev}, 1, $buf);

    $buf = $s->converse(read_bytes=>5);
    $s->debug("After reset: %s\n", $s->hex($buf));
    my @buf = unpack('C*', $buf);
    if ($buf[0] != 0x91 || $buf[2] != 0x92 || $buf[3] != 0x04) {
        die "$s->{tcename} doesn't respond properly as an HLB VBus device";
    }
    my $version = sprintf("%d.%d", $buf[4] >> 4, $buf[4] & 0x0F);
    return $version;
}

sub read_avt {
    my ($s, $len) = @_;
    my $buf = '';
    my $toget = $len;
    local $SIG{ALRM} = sub {
        die(sprintf "read_avt: Timeout reading byte %d of $toget: got %s",
            length($buf), $s->hex($buf))
    };
    alarm 10;
    while (length($buf) < $len) {
        sysread($s->{dev}, $buf, $len - length($buf), length($buf))
            or last;
    }
    alarm 0;
    length($buf) == $len
        or die(sprintf "read_avt: Error reading byte %d of $len: got %s",
            length($buf), $s->hex($buf));
    return $buf;
}

sub write_avt {
    my ($s, $buf) = @_;
    $s->debug("write_avt %d bytes: %s\n", length($buf), $s->hex($buf));
    my $wrote = syswrite($s->{dev}, $buf);
    my $len = length($buf);
    $wrote == $len
        or die "Error writing $len bytes, write() returned $wrote: $!";
    return $len;
}

sub converse {
    my ($s, %args) = @_;
    if (defined $args{write}) {
        $s->debug("converse write %d bytes: %s...\n",
            length($args{write}), $s->hex($args{write}));
        $s->write_avt($args{write});
    }
    my $buf;
    return $buf if $s->{no_device};
    if (defined $args{read_bytes}) {
        $s->debug("converse read $args{read_bytes} bytes...\n");
        $buf = $s->read_avt($args{read_bytes});
        $s->debug("converse read %d bytes - %s\n",
            length($buf), $s->hex($buf));
    }
    my $match = $args{match};
    if (defined $match) {
        die "Read too few bytes for match"
            if length($buf) < length($match);
        if ($match ne substr($buf, 0, length($match))) {
            die(sprintf "converse: bad response, expected %s, got %s",
                $s->hex($match), $s->hex(substr($buf, 0, length($match))));
        }
        $s->debug("converse matched with %s\n", $s->hex($match));
    }
    return $buf;
}

sub read_block {
    my ($s, $addr, $len) = @_;

    $s->debug("read_block, addr=0x%.04X, len=$len\n", $addr);
    my $req = pack('CCVV', 0x59, 0x37, $addr, $len);
    $s->converse(write=>$req, read_bytes=>2, match=>"\x61\x37");
    $s->debug("read_block, reading $len bytes...\n");
    my $buf = $s->read_avt($len);
    $s->debug("read_block, got %d bytes.\n", length($buf));
    return $buf;
}

sub reflash {
    my ($s, $hexfh, $file_size) = @_;

    $s->{line_num} = 0;
    $s->init_progress($file_size);
    my $file_read = 0;
    my $next_progress = 0;

    # Reset and check version
    unless ($s->{no_device}) {
        $s->debug("reflash, reset and check version\n");
        my $version = $s->reset;
        $s->progress("$s->{tcename}: Previous firmware version $version");
    }

    # Enter FLASH mode
    $s->debug("reflash, enter FLASH mode\n");
    $s->converse(write=>"\xE1\x96", read_bytes=>5, match=>"\x91\x2D\x92\x04");

    if ($s->{erase}) {
        $s->debug("reflash, erasing\n");
        my $addr = 0x8000; # Don't erase the boot block;
        $addr += 0x20000 if $s->{test_mode};
        while ($addr < 0x40000) {
            $s->progress("Erase 0x%x", $addr);
            my $cmd = pack("CCN", 0x55, 0x38, $addr);
            $s->converse(write=>$cmd, read_bytes=>3, match=>"\x92\x2B\x80");
            $addr += 0x8000;
        }
    }

    while (defined (my $line = <$hexfh>)) {
        ++$s->{line_num};
        $file_read += length($line);
        #$line =~ s/\s+$//;
        if ($line !~ /^S(\d)((?:[0-9a-fA-F]{2})+)\s*$/s) {
            die "Bad format for 'S' record file: $line";
        }
        my ($rec_type, $hex) = ($1, $2);
        next if $rec_type =~ /^[05789]$/; # Ignore these record types
        die "Unsupported S record: $line" unless $rec_type =~ /^[123]$/;

        my $addr_count = $rec_type + 1;

        my $buf = pack('H*', $hex);

        my $checksum = ord(substr($buf, -1, 1, ''));
        my ($sum) = unpack('%8C*', $buf);
        $sum = 0xFF - $sum;
        die(sprintf "Bad checksum 0x%.02X, should be 0x%.02X\n%s",
            $sum, $checksum, $line)
            if $sum != $checksum;
#printf "OK Checksum 0x%.02X, should be 0x%.02X\n", $sum, $checksum;

        my ($byte_count, $payload) = unpack('Ca*', $buf);
        $s->debug("line: ($line)\n");
        $s->debug("payload: %s\n", $s->hex($payload));

        # Pad front of address
        substr($payload, 0, 0, "\x00" x (4 - $addr_count));
        my ($addr, $data) = unpack('Na*', $payload);

        $s->debug("line: ($line)\n");
        $s->debug("addr=0x%.04X, data: %s\n", $addr, $s->hex($data));

        if ($s->{test_mode}) {
            $addr += 0x20000;
        }

        if ($s->{do_write}) {
            $s->debug("do_write, addr=0x%.04X, data: %s\n",
                $addr, $s->hex($data));
            my $cmd = pack('CCNCa*', 0x56, 0x3A, $addr, length($data), $data);
            $s->debug("cmd: %s\n", $s->hex($cmd));
            $s->debug("do_write, addr=0x%.04X, data: %s\n",
                $addr, $s->hex($data));
            $s->converse(write=>$cmd, read_bytes=>3, match=>"\x62\x3A\x80");
        }

        # KLUDGE we can't do this because the board will reset on the
        # command after the memory dump if we try - no idea why
        if (0 && $s->{verify}) {
            my $cmd = pack('CCNN', 0x59, 0x37, $addr, length($data));
            my $cmp = pack('CCa*', 0x61, 0x37, $data);
            $s->debug("verify, addr=0x%.04X, data: %s\n",
                $addr, $s->hex($data));
            $s->debug("verify cmd: %s\n", $s->hex($cmd));
            $s->debug("verify cmp: %s\n", $s->hex($cmp));

            $s->converse(write=>$cmd, read_bytes=>length($cmp), match=>$cmp);
        }
        if ($file_read >= $next_progress) {
            my $steps = 100.0;
            $next_progress = int(($file_read / $file_size) * $steps) + 1;
            $s->progress_update($file_read);
            $s->progress("%s", $s->{verify}
                ? "Write / Verify FLASH" : "Write FLASH");
            $next_progress *= $file_size / $steps;
        }
    }
    my $version = $s->reset;
    $s->progress_update($file_size);
    $s->progress("$s->{tcename} Finished, firmware version now $version\n");
}

1;
