use strict;

package SoMat::eDAQ::Structurizer;

use vars qw($VERSION);
$VERSION = sprintf("%d.%02d", q$Revision: GIT $ =~ /(\d+)\.(\d+)/);

# TODO
# Allow 'typedef' as argument - don't need n objects for this
# Allow 'packed' to be from a file/offset - read and write properly
# 

my $Symbols = {};

sub new {
    my $class = shift;
    my $s = bless {}, ref $class || $class;
    return $s->initialize(@_);
}

sub initialize {
    my ($s, $init) = @_;
    $s->{stack} = [];
    if (ref $init) { # File Handle
        $s->{symtype} = undef;
        $s->{packed} = "";
        $s->{fh} = $init;
    } elsif (defined $init) { # Symbol
        $s->{symtype} = $s->get_record($init);
        $s->clear;
    } else {
        $s->{packed} = "";
    }
    return $s;
}

sub _get_symbols {
    return $Symbols;
}

sub get_record {
    my ($s, $symtype) = @_;
    $symtype = $s->{symtype} unless defined $symtype;
    return $symtype if ref($symtype) eq 'HASH';
    return $Symbols->{$symtype} if exists $Symbols->{$symtype};
    (my $fname = $symtype) =~ s/\s/-/g;
    my $r = do "SoMat/eDAQ/Structurizer/$fname.pl";
    return undef unless defined $r;
    $Symbols->{$symtype} = $r;
    return $r;
}

sub set_record { # KLUDGE Note: NEVER USED
    my ($s, $symtype, $r) = @_;
    $Symbols->{$symtype} = $r;
}

sub push_type {
    my ($s, $symtype) = @_;
    push(@{$s->{stack}}, [$s->{symtype}, $s->{packed}, $s->{fh}, $s->{fpos}]);
    $s->{symtype} = undef;
    $s->{packed} = "";
}

sub pop_type {
    my ($s) = @_;
    my $popped = pop(@{$s->{stack}});
    $popped = [] unless defined $popped;
    ($s->{symtype}, $s->{packed}, $s->{fh}, $s->{fpos}) = @$popped;
}

sub _descriptor {
    my $s = shift;
    #my $symtype = shift;
    my $symtype = $s->{symtype} || shift;
    my $sym = $s->get_record($symtype);
    my $off = 0;
    my $size = 0;
    my $pack;
    while (@_) {
        my $tag;
        if (exists $sym->{count}) {
            $tag = shift;
#print STDERR "Array, tag = $tag\n";
            if ($tag !~ /^\d+$/) {
                warn "Bad index for array: $tag\n";
                return undef;
            }
            $off += $tag * $sym->{basesize};
            $size = $sym->{basesize};
            $sym = $sym->{type};
        } elsif (exists $sym->{struct} || exists $sym->{union}) {
            $tag = shift;
#print STDERR "Struct, tag = $tag\n";
            my $new_sym;
            my $sref = $sym->{struct} || $sym->{union};
            for my $e (@$sref) {
                if ($tag eq $e->{name}) {
                    $new_sym = $e;
                }
            }
            unless (defined $new_sym) {
                warn "Bad element for struct: $tag\n";
                return undef;
            }
            $sym = $new_sym;
            $off += $sym->{offset};
            $size = $sym->{size};
        } elsif (exists $sym->{type}) {
#print STDERR "Type, type = ", $sym->{type}, "\n";
            $sym = $sym->{type};
            $sym = $s->get_record($sym) unless ref $sym eq 'HASH';
        } else {
            $tag = shift;
#print STDERR "BUSTED: tag = $tag\n";
        }
    }
    $pack = $s->_get_pack($sym);
    #warn("No pack for " . $sym->{name}) unless defined $pack;
#print STDERR "Done, pack = $pack\n";
    return ($pack, $off, $size);
}

sub get_raw {
    my $s = shift;
    return $s->{packed} unless @_;
    my ($pack, $off, $size) = $s->_descriptor(@_);
    return substr($s->{packed}, $off, $size);
}

sub get_size {
    my $s = shift;
    my (undef, undef, $size) = $s->_descriptor(@_);
    return $size;
}

sub get {
    # KLUDGE same as above code - avoid yet another method call...
    my $s = shift;
    my ($pack, $off, $size) = $s->_descriptor(@_);
    my $packed = substr($s->{packed}, $off, $size);
    unless (defined $pack) {
        warn("No pack for " . join('.', @_[0..$#_]));
        return $packed;
    }
    return unpack($pack, $packed);
}

sub put {
    my $s = shift;
    my $val = pop @_;
    my ($pack, $off, $size) = $s->_descriptor(@_);
    return unless defined $pack;
    my $packed = pack($pack, $val);
    if ($off + $size > length($s->{packed})) {
        die "Structurizer put(" . join(', ', @_) . ", val=$val) symtype=$s->{symtype}, pack=$pack, off=$off, size=$size OUT OF BOUNDS";
    }
    substr($s->{packed}, $off, $size, $packed);
}

sub size {
    my ($s, $symtype) = @_;
    $symtype = $s->{symtype} unless defined $symtype;
    my $sym = $s->get_record($symtype);
    return $sym->{size} if defined $sym && exists $sym->{size};
    return $symtype;
}

sub read {
    my ($s, $symtype, $fpos) = @_;
    if (defined $fpos) {
        $s->{fpos} = $s->{fh}->seek($fpos, 0); # KLUDGE errors
    } else {
        # KLUDGE Shouldn't do this if we know...
        #$s->{fpos} = $s->{fh}->seek(0, 1); # KLUDGE errors
    }
    my $size = $symtype;
    $size = $s->size($symtype) unless $symtype =~ /^\d/;
print STDERR "UNDEF symtype!\n" unless defined $symtype;
print STDERR "UNDEF size!\n" unless defined $size;
    $s->{fh}->read($s->{packed}, $size); # KLUDGE check errors
    $s->{symtype} = $symtype;
#print "read got ", length($s->{packed}), " bytes\n";
    return $s->{packed};
}

sub new_sym {
    my ($s, $symtype) = @_;
    $s->{symtype} = $symtype; # KLUDGE get_record?
    $s->clear;
    return $s->{packed};
}

sub clear {
    my ($s) = @_;
    $s->{packed} = "\0" x $s->size;
}

sub write {
    my ($s) = @_;
    return undef unless $s->{fh}->seek($s->{fpos}, 0) == $s->{fpos};
    return ($s->{fh}->write($s->{packed}) == length($s->{packed}));
}

sub _get_pack {
    my ($s, $sym) = @_;
    return $sym->{pack} if exists $sym->{pack};
    if (exists $sym->{type}) {
#print "XXX _get_pack($sym->{type})\n";
        return $s->_get_pack($sym->{type}) if ref $sym->{type} eq 'HASH';
        if (exists $sym->{struct}) {
#print "XXX _get_pack($sym->{type}) DOING STRUCT\n";
            my $pack = "";
            for my $e (@{$sym->{struct}}) {
                $pack .= $e->{pack};
            }
#print "XXX _get_pack($sym->{type}) STRUCT returning '$pack'\n";
            return $pack;
        }
#print "XXX _get_pack($sym->{type}) doing get_record\n";
        return $s->_get_pack($s->get_record($sym->{type}));
    }
    return undef;
} 

sub _dump {
    my $s = shift;
    my ($sym, $name, $data, $off) = @_;
    if (exists $sym->{count}) {
        my $ss = $s->get_record($sym->{type});
        $off += $sym->{offset};
        for my $i (0 .. $sym->{count} - 1) {
            $s->_dump($ss, "${name}\[$i\]", $data, $off);
            $off += $sym->{basesize};
        }
        return;
    }
    if (exists $sym->{offset}) {
        $off += $sym->{offset};
    } else {
        warn("NO OFFSET for $sym->{name} in $name\n")
            unless exists $sym->{type}; # No offset for array element 
    }
    if (exists $sym->{struct}) {
        for my $ss (@{$sym->{struct}}) {
            $s->_dump($ss, $name . "." . $ss->{name}, $data, $off);
        }
    } elsif (exists $sym->{union}) {
        print "UNION _dump\n";
    } elsif (exists $sym->{pack}) {
        my $pack = $s->_get_pack($sym);
        unless (defined $pack) {
            warn "No pack for " . $sym->{name} . "\n";
            return undef;
        }
        print "$name ($pack @ $off) = ";
        print unpack($pack, substr($data, $off)), "\n";
    } elsif (exists $sym->{type}) {
        $s->_dump($s->get_record($sym->{type}), $name, $data, $off);
    }
}

sub dump {
    #my ($s, $symtype, $name) = @_;
    my $s = shift;
    my $symtype = $s->{symtype} || shift;
    my $name = shift;
    $s->_dump($s->get_record($symtype), $name, $s->{packed}, 0);
}

1;

