
use strict;

package SoMat::eDAQ::Web;
require SoMat::eDAQ::WebServer;
@SoMat::eDAQ::Web::ISA = qw(SoMat::eDAQ::WebServer);
require Symbol;
require SoMat::ConfigFile;

sub contact_support_string {
    my ($s) = @_;
    return "<p>Please contact " .
        $s->link('SoMat Customer Support', 'mailto:support@somat.com') .
        " with the above messages and the contents of the " . 
        $s->link("eDAQ LogBook", $s->ui_url('Test/Logbook')) .
        " for assistance.</p>";
}

sub need_notification_support_string {
    my ($s) = @_;
    $s->{need_notification_support_string} = 1;
}

sub maybe_print_notification_header {
    my ($s) = @_;
    delete $s->{need_notification_support_string};
    unless ($s->{have_printed_notification_header}) {
        print qq{<div class="alert">\n};
        $s->{have_printed_notification_header} = 1;
    }
}

sub maybe_print_notification_footer {
    my ($s) = @_;
    if ($s->{have_printed_notification_header}) {
        if ($s->{need_notification_support_string}) {
            print $s->contact_support_string();
        }
        print "</div>\n";
        delete $s->{have_printed_notification_header};
    }
}

sub all_files_in {
    my ($dir) = @_;
    local(*DIR);
    opendir(DIR, $dir);
    my @files = grep { -f $_ } map { "$dir/$_" } readdir(DIR);
    closedir(DIR);
    return @files;
}

sub print_notifications {
    my ($s) = @_;
    my $pid = SoMat::ConfigFile->read_pid_file("/var/run/edaqserver.pid");
    my $running = ($pid && kill(0, $pid));

    if (-f "/sram/.early_edaqserver_failure") { 
        $s->maybe_print_notification_header();
        print "<p>The eDAQ has encountered an error in the early phases
               of its initialization and has been restarted.  To allow
               a firmware update, the eDAQ's test-processing software
               was not started on this boot.</p>";
        $s->need_notification_support_string();
    } elsif (-f "/var/state/powerd_watchdog_failed") {
        $s->maybe_print_notification_header();
        print "<p>The eDAQ was unable to turn on the Power
               Controller's watchdog interface.  Starting a test is
               impossible at this time.  The eDAQ will keep retrying
               the Power Controller command, and will reboot
               immediately if it succeeds.  Otherwise, the eDAQ should
               be restarted by unplugging main power, waiting for the
               unit to power off, and then reapplying power.</p>";
        $s->need_notification_support_string();
    } elsif (!$running && !-f "/var/state/edaqserver-dead-on-purpose") {
        $s->maybe_print_notification_header();
        print "<p>The eDAQ's test-processing software is not running.</p>";
        $s->need_notification_support_string();
        if (-f "/sram/.early_edaqserver") {
            print "<p>The eDAQ's test software did not finish
                   initializing.</p>";
        }
    } elsif (-f "/sram/.early_edaqserver") {
        $s->maybe_print_notification_header();
        print "<p>The eDAQ's test software has not yet finished
               initializing.  If this state persists for more than a
               few minutes, it is possible that the software is not
               responding.</p>";
        $s->need_notification_support_string();
    }
    $s->maybe_print_notification_footer();
    my @permanent_notes = all_files_in("/eDAQ/web_notifications.d");
    my @temporary_notes = all_files_in("/var/state/web_notifications.d");
    for my $f (@permanent_notes, @temporary_notes) {
        if (-f $f) {
            local($/, *FH);
            open(FH, $f);
            print <FH>;
            close FH;
        }
    }
}

sub is_text {
    my ($s) = @_;
    return 1 if $s->get('HTTP_Path') =~ /\.txt$/;
    return 0;
}

sub is_html {
    my ($s) = @_;
    return !$s->is_text;
}

sub edaq_header {
    my ($s, $title, $tab, %args) = @_;

    if ($s->is_text) {
        $s->send_header;
        return 1;
    }
    $title = $s->{title} || $s->{print_title} unless defined $title;
    $s->{title} = $title;
    my $path = $s->get('HTTP_Path');
    $path =~ s/^\/eDAQ\///;
    my @path = split(/\//, $path);
    $s->{http_path} = [ @path ];
    $tab = shift @path unless defined $tab;
    $s->{current_section} = $tab;

    $s->new_header($title, %args);
    $s->test_status_bar();
    $s->print_notifications();

    print qq{<div class="page-content">};
    $s->tab_bar($tab); # KLUDGE get from path if undef
    print qq{<div id="contents">};

    my $cur_url = "/eDAQ/$tab";
    my @chain;
    while (0 && @path) { # KLUDGE Don't do this for now
        my $link = shift @path; 
        $cur_url .= "/$link";
        $link =~ s/_/ /g;
        push @chain, (@path) ? $s->link($link, $cur_url) : $link;
    }
    my $chain = join(' &raquo; ', @chain);
    print qq(<b>$chain</b><p>) if $chain;
    print "\n";
    print "<h2>$title</h2>\n" if $s->{print_title};
    return 1;
}

sub get_firmware_version {
    my $f;
    open $f, "/etc/version/string";
    my $v = <$f>;
    close $f;
    chomp $v;
    $v ||= 'UNRELEASED';
    $v .= " DIRTY" if (-s "/etc/version/git-dirty");
    return $v;
}

sub is_edaq_classic {
    return -f '/eDAQ/.is-an-edaq-classic';
}

sub is_edaq_plus {
    return -f '/eDAQ/.is-an-edaq-plus';
}

sub is_edaq_lite {
    return -f '/eDAQ/.is-an-edaq-lite';
}

sub edaq_footer {
    my ($s) = @_;
    return 0 if $s->is_text;
    my ($version) =
        $s->link(get_firmware_version, $s->ui_url('System/Version'));

    if (exists $s->{current_section}) { # We have tabs
        print <<"EOP";
<br>
</div>
</div>
EOP
    }

    print <<"EOP";
<div class="footer">
eDAQ Firmware $version Copyright &copy; 1999 &ndash; 2009 <a href="http://www.somat.com">SoMat Corporation</a>. All Rights Reserved.
</div>
</body>
</html>
EOP
    return 0;
}

sub canceled {
    my ($s, $default) = @_;
    return 0 unless defined $s->param('_Cancel');
    my $ret = $s->param('_Return');
    unless ($ret) {
        if ($default) {
            $ret = $s->ui_url($default);
        } else {
            # KLUDGE maybe look for Referer as well?
            $ret = $s->script_path;
            $ret =~ s/\/\w+$//;
        }
    }
    $s->redirect($ret);
    return 1;
}

sub error_box {
    my ($s, $prefix, $message) = @_;
    ($message, $prefix) = ($prefix, 'Error') unless defined $message;
    my $tab = $s->{current_section} || '';
    print qq{<div class="errorbox">\n}
        . qq{<p><b>$prefix:</b> $message</p>\n};
    print qq{<p>Return to } . $s->link($tab, $s->ui_url($tab)) . qq{</p>\n}
        if $tab;
    print qq{</div>\n};
}

sub assure_no_test_running {
    my ($s) = @_;
    my $status = $s->test_status;
    if ($status->{RunStarted} || $status->{RunPreview}) {
        $s->edaq_header("ERROR: $s->{title}");
        print qq{<h2>$s->{title}</h2>\n};
        $s->error_box(qq{This action ($s->{title}) is not allowed while}
            . qq{ a test is running on the eDAQ.</p>\n});
        $s->edaq_footer;
        die 'PAGE_ABORT';
    }
    return 0;
}

sub new_header {
    my ($s, $title, %args) = @_;

    $args{head_extra} ||= '';

    $s->send_header('text/html');

    my $icon = ''; # '<link rel="SHORTCUT ICON" href="/favicon.ico">';

# Don't use strict until we fix it to work...
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
    print <<"EOP";
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>$title</title>
$icon
<link rel="stylesheet" type="text/css" href="/style.css" media="all">
$args{head_extra}
</head>
<body class="$s->{current_section}">
EOP
    return 0;
}


sub tab_bar {
    my ($s, $current) = @_;
    my @tabs = qw(System Hardware Channels Test Data Custom Help);
    unless (defined $current) {
        # KLUDGE get $current from path here
    }
    my $bg = "#FF0000";
    my $fg = "#FFFFFF";
    my $sbg = "#FFC0C0";
    my $cols = @tabs * 2  + 1;

if (0) {
    print <<"EOP";
<center>
<table border=0 cellspacing=0 cellpadding=0 width="100%">
<tr>
EOP

    for my $tab (@tabs) {
        print qq(<td width=4>&nbsp;</td>\n);
        my $ttab = $tab;
        my $tbg = $sbg;
        my $color = "";
        my $style = 'tab';
        if ($tab eq $current) {
            $color = " color=$fg";
            $tbg = $bg;
            $style = 'tab-current';
        }
        (my $ttext = $tab) =~ s/^.*\///;
        $ttext =~ s/_/ /g;
        $ttab = $s->link($ttext, "/eDAQ/$tab");
        print qq(<td bgcolor=$tbg align=center nowrap>);
        print qq(<span style="$style">$ttab</span>);
        print qq(</td>\n);
    }

    print <<"EOP";
<td width=4>&nbsp;</td>
</tr>
<tr>
<td colspan=$cols bgcolor=$bg><img width=1 height=2 alt=""></td>
</tr>
</table>
</center>
EOP
} # if 0

    print qq{<ul id="menu">\n};
    my $i = 1;
    for my $tab (@tabs) {
        print qq{<li id="nav-$i"><a href="/eDAQ/$tab">$tab</a></li>\n};
        ++$i;
    }
    print qq{</ul>\n};

    return 0;
}

sub header {
    my ($s, $title) = @_;

    $s->send_header('text/html');

    print <<"EOP";
<html>
<head>
<title>$title</title>
<link rel="stylesheet" type="text/css" href="/style.css" media="all">
</head>
<body>
EOP

    return 0;
}

# KLUDGE replace with read_config_file or similar when ConfigFile.pm is done
sub read_value_from_file {
    my ($file) = @_;
    my $fh = Symbol::gensym();
    open($fh, $file);
    my $retval = <$fh>;
    close $fh;
    chomp $retval;
    return $retval;
}

sub test_status_bar {
    my ($s, %args) = @_;
    my $c = $s->test_status(%args) if $s->get('HTTP_Port') == 80;
    my $status = '&nbsp;';
    my $controls = '&nbsp;';
    if (defined $args{status}) {
        $status = $args{status};
    } elsif (!$c) {
        # Do nothing
    } elsif ($c->{TestInitialized}) {
        # \273 => &raquo;
        $status = "Test: " . $c->{SetupName} . " \&raquo\; ";
        my $dimen = qq(border="0" width="16" height="16");
        my $ret = "?_Return=" . $s->get('HTTP_Path');
        if ($c->{RunStarted}) {
            $status .= "Run " . $c->{Run} . " &raquo; " . $c->{RunTime};
            my $img = qq(<img src="/images/stop.gif" alt="Stop Test" $dimen>);
            $controls = $s->link($img, "/eDAQ/Test/StopRun$ret");
        } else {
            $status .= "Next Run: " . $c->{Run};
            my $img = qq(<img src="/images/start.gif" alt="Start Test" $dimen>);
            $controls = $s->link($img, "/eDAQ/Test/StartRun$ret");
        }
    } else {
        $status = "No Test Initialized";
    }
    
    my @t = localtime;
    my $address = $args{edaq}
        || $s->get('HTTP_Host') || $s->get('HTTP_Address');
    my $time = sprintf "%.04d-%.02d-%.02d %.02d:%.02d:%.02d",
        $t[5] + 1900, $t[4] + 1, $t[3], $t[2], $t[1], $t[0];
    
    my $batt = '&nbsp;';
    my $batt_status = '&nbsp;';
    unless ($s->is_edaq_classic()) {
        my $charging = read_value_from_file("/var/state/power_charging");
        my $charge_level = read_value_from_file("/var/state/power_charge_level");
        my $batt_number = 0;
        my $batt_alt = "Empty/Not present";
        my $batt_charge = ($charging ? "charge" : "");
        
        if ($charge_level >   0) { $batt_number = 1; $batt_alt = "Low"; }
        if ($charge_level > 400) { $batt_number = 2; $batt_alt = "Okay"; }
        if ($charge_level > 500) { $batt_number = 3; $batt_alt = "High"; }
        
        my $batt_icon = "/images/batt$batt_number$batt_charge.gif";
        my $batt_dimen = qq(width="32" height="16");
        $batt_alt = "$batt_alt" . ($charging ? ", Charging" : "");
        $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);

        #
        # Per Brad, the following is how we determine whether or not the
        # battery pack is good or bad.  The values given for current and
        # voltages are approximate and may require "tweaking".
        #
        # Battery condition can only be determined if drain on the battery is
        # below the "cold crank" threshold.  If the battery drain is above the
        # cold cranking threshold then determination of the battery pack
        # condition rests on the whether or not the charger is on and what the
        # voltage at the battery is.  If the charger is on, the voltage should
        # be above the nominal battery pack voltage.  The minimum voltage level
        # for a good battery pack when the charger is on is ~13V.  If the
        # charger is off and the battery pack voltage is above ~11V then the
        # battery pack is good.  We put something on the web page only when the
        # battery pack is bad.
        #
        my $sense_ma = read_value_from_file("/var/state/power_sense_ma");
        if (scalar($sense_ma) >= -200) {
            my $millivolts = read_value_from_file("/var/state/power_nicad_mv");
            if (scalar($charging) > 0) {
                if (scalar($millivolts) < 11400) {
                    $batt_status = "Bad Battery";
                    $batt_icon = "/images/batt1.gif";
                    $batt_dimen = qq(width="32" height="16");
                    $batt_alt = $batt_status;
                    $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);
                } elsif (scalar($millivolts) < 12600) {
                    $batt_status = "Marginal Battery";
                    $batt_icon = "/images/batt2.gif";
                    $batt_dimen = qq(width="32" height="16");
                    $batt_alt = $batt_status;
                    $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);
                }
            } else {
                if (scalar($millivolts) < 11000) {
                    $batt_status = "Bad Battery";
                    $batt_icon = "/images/batt1.gif";
                    $batt_dimen = qq(width="32" height="16");
                    $batt_alt = $batt_status;
                    $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);
                } elsif (scalar($millivolts) < 12000) {
                    $batt_status = "Marginal Battery";
                    $batt_icon = "/images/batt2.gif";
                    $batt_dimen = qq(width="32" height="16");
                    $batt_alt = $batt_status;
                    $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);
                }
            }
        } else {
            # If there is any outflow of current from the battery pack (over
            # the noise level from the ADC) we tell the user that there is
            # something wrong with the battery charging/monitoring circuit.
            # This can include problems with one or any of battery pack,
            # battery charger, battery monitoring circuit.
            #
            # However, there are times when large current outflows from the
            # battery pack are OK and indicating that there is a
            # charging/monitoring problem will not be entirely accurate.
            #
            # The only time we should ever get ligitimate large outflows of
            # current from the battery is when the eDAQ is being brought down
            # as a result of a power loss OR under "cold crank" conditions.
            # In first case, the user is unlikely to every update the webpage
            # while the eDAQ is going down, so it's unlikely that the user
            # would ever see the bad charging indication.  Likewise, for "cold
            # crank" conditions, the condition is very short lived and the user
            # will likely never see bad charging indication under those
            # circumstances as well.
            #
            # There is no way to predict the current draw from the battery pack
            # under "cold crank" conditions so we will just have to live with
            # it the way it is.
            $batt_status = "Charging Problem";
            $batt_icon = "/images/batt1.gif";
            $batt_dimen = qq(width="32" height="16");
            $batt_alt = $batt_status;
            $batt = qq(<img src="$batt_icon" alt="[Battery: $batt_alt]" $batt_dimen>);
        }
    }

#     print <<"EOP";
# <div class="status-bar">
# <table border=0 cellspacing=0 cellpadding=0 width="100%">
# <tr>
# <td width="64" align="left"><img src="/images/edaq16.gif" alt="eDAQ" border="0" width="64" height="16"></td>
# <td width="1%" align="left" nowrap>&nbsp;&nbsp;$status&nbsp;&nbsp;</td>
# <td width="1%" align="left" nowrap>$controls</td>
# <td width="97%" align="right" nowrap>$batt</td>
# <td width="1%" align="right" nowrap>&nbsp;&nbsp;$time</td>
# </tr>
# </table>
# </div>
# EOP

    print <<"EOP";
<div class="status-bar">
<span class="sb-test">
<span class="sb-logo"><img src="/images/eDAQ_tiny.gif" alt="eDAQ" border="0" height="16" width="46"></span>
<span class="sb-test-status">$status</span>
<span class="sb-test-control">$controls</span>
</span>
<span class="sb-status">
<span class="sb-batstatus">$batt_status</span>
<span class="sb-battery">$batt</span>
<span class="sb-node">$address</span>
<span class="sb-date">$time</span>
</span>
</div>
EOP

    return 0;
}

sub edaq_sif_file {
    my ($s) = @_;
    my $disk = '/hd/eDAQ/sif00000.sic';
    my $sram = 'sram:/data/test.sif';
    return $sram unless -f $disk;
    return $sram unless $s->test_status->{SifFileOnDisk};
    return $disk;
}

sub sif_file_name {
    my ($s) = @_;
    (my $name = $s->test_status->{SetupFileLong}) =~ s/\.tce$/\.sif/;
    return $name;
}

sub ui_url {
    my ($s, $url) = @_;
    return "/eDAQ/$url";
}

sub link {
    my ($s, $text, $url) = @_;
    $url = $text unless defined $url;
    $url = $s->escape_html($url);
    $url =~ s/ /%20/g;
    return qq(<a href="$url">$text</a>);
}

sub rlink {
    my ($s, $text, $url) = @_;
    return $s->link($text, $s->script_path . "/$url");
}

sub table {
    my $s = shift;
    my $t = SoMat::eDAQ::Web::Table->new(@_);
    $t->{web} = $s;
    return $t;
}

sub form {
    my $s = shift;
    my $f = SoMat::eDAQ::Web::Form->new(@_);
    $f->{web} = $s;
    return $f;
}

sub ruled_shaded_table {
    my ($s, $args) = _arghash(@_);
    my $defaults = {
        #'rulewidth'=> 1,
        #'row_bg' => ['#C0C0C0', ['#FFFFFF', '#E0E0E0']]
        even_odd_shaded=>1,
    };
    my $t = $s->table($s->_arg_merge($args, $defaults));
    return $t;
}

# Generic argument processing stuff

sub _arghash {
    my $s = shift;
    my $args = (defined $_[0] && ref($_[0]) eq "HASH") ? $_[0] : { @_ };
    return ($s, $args);
}

sub _arg_merge {
    my $s = shift;
    my $arg = shift;
    while (@_) {
        my $merge = shift @_;
        while (my ($key, $val) = each %$merge) {
            $arg->{$key} = $val unless exists $arg->{$key};
        }
    }
    return $arg;
}

sub _argvalues {
    my $s = shift;
    my $args = shift;
    if (ref($_[0]) eq 'HASH') { # Process defaults
        my $defaults = shift;
        $args = $s->_arg_merge($args, $defaults);
    }
    my @a = ();
    while (defined (my $name = shift)) {
        next unless exists $args->{$name};
        my $v = $args->{$name};
        $v = '' unless defined $v;
        $v = $s->escape_html($v);
        push(@a, qq{$name="$v"});
    }
    unshift(@a, "") if @a; # Leading space if @a not empty
    return join(" ", @a);
}

sub write_spreadsheet {
    my ($s, $file, $prefix) = @_;

    my $fh = Symbol::gensym();
    return undef unless open($fh, ">$file");
    my $row = 0;
    my $col = 0;
    while (1) {
        my $p = $s->param("$prefix$row\:$col");
        if (defined $p) {
            print $fh "\t" if $col++;
            print $fh $p;
        } elsif ($col) {
            print $fh "\n";
            $col = 0;
            ++$row;
        } else {
            last;
        }
    }
    return close($fh);
}

# Forms stuff
package SoMat::eDAQ::Web::Form;
@SoMat::eDAQ::Web::Form::ISA = qw(SoMat::eDAQ::Web);

*_arghash = \&SoMat::eDAQ::Web::_arghash;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    return $s;
}

sub param {
    my $s = shift;
    return undef unless exists $s->{web};
    return $s->{web}->param(@_);
}

sub _value {
    my ($s, $args, $valuetag) = @_;
    $valuetag = 'value' unless defined $valuetag;
    my $val = undef;
    if (exists $args->{$valuetag}) {
        $val = $args->{$valuetag};
    } else {
        $val = $s->param($args->{name});
        if (!defined $val && exists $args->{default}) {
            $val = $args->{default};
        }
    }
    return $val;
}

sub _checked {
    my ($s, $args) = @_;
    my $checked = 0;
    if (exists $args->{checked}) {
        ++$checked if $args->{checked};
    } else {
        my $v = $s->param($args->{name});
        if (defined $v) {
            ++$checked if $args->{value} eq $v;
        } else {
            ++$checked if $args->{default};
        }
    }
    return ' checked' if $checked;
    return '';
}

sub begin {
    my ($s, $args) = _arghash(@_);
    $args->{method} = 'GET' unless exists $args->{method};
    $args->{action} = $s->{web}->script_path unless exists $args->{action};
    my $p = $s->_argvalues($args, 'method', 'action', 'enctype');
    return qq(<form$p>\n);
}

sub begin_multipart {
    my $s = shift;
    return $s->begin(@_, 'method'=>'POST', 'enctype'=>'multipart/form-data');
}

# Form elements

sub checkbox {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'value');
    $p .= $s->_checked($args);
    return qq(<input type="checkbox"$p>);
}

sub filefield {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'size', 'value');
    return qq(<input type="file"$p>);
}

sub hidden {
    my ($s, $args) = _arghash(@_);
    #$args->{value} = $s->param($args->{name}) unless exists $args->{value};
    my $v = $s->_value($args);
    $args->{value} = $v if defined $v;
    my $p = $s->_argvalues($args, 'name', 'value');
    return qq(<input type="hidden"$p>);
}

sub image {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'src', 'align');
    return qq(<input type="image"$p>);
}

sub password {
    my ($s, $args) = _arghash(@_);
    #$args->{value} = $s->param($args->{name}) unless exists $args->{value};
    my $v = $s->_value($args);
    $args->{value} = $v if defined $v;
    my $p = $s->_argvalues($args, 'name', 'size', 'maxlength', 'value');
    return qq(<input type="password"$p>);
}

sub radio {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'value');
    $p .= $s->_checked($args);
    return qq(<input type="radio"$p>);
}

sub reset {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'value');
    return qq(<input type="reset"$p>);
}

sub select {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'size'); # KLUDGE do 'multiple'
    #my $selected = (exists $args->{value})
    #    ? $args->{value} : $s->param($args->{name});
    my $selected = $s->_value($args);
    my $r = qq(<select$p>);
    if (exists $args->{optpairs}) {
        my @opt = @{$args->{optpairs}};
        while (@opt) {
            my $string = shift @opt;
            my $name = shift @opt;
            my $sel = "";
            $sel = qq( selected="1") if $name eq $selected;
            $r .= qq(<option value="$name"$sel>$string);
        }
    } elsif (exists $args->{opts}) {
        for my $name (@{$args->{opts}}) {
            my $sel = "";
            $sel = qq( selected="1") if $name eq $selected;
            $r .= qq(<option$sel>$name);
        }
    }
    $r .= qq(</select>);
    return $r;
}

sub submit {
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'value');
    return qq(<input type="submit"$p>);
}

sub textarea { # KLUDGE!
    my ($s, $args) = _arghash(@_);
    my $p = $s->_argvalues($args, 'name', 'rows', 'cols');
    my $val = $s->escape_html($s->_value($args) || "");
    return qq(<textarea$p>$val</textarea>);
}

sub textfield {
    my ($s, $args) = _arghash(@_);
    #$args->{value} = $s->param($args->{name}) unless exists $args->{value};
    my $v = $s->_value($args);
    $args->{value} = $v if defined $v;
    my $p = $s->_argvalues($args, 'name', 'size', 'maxlength', 'value');
    return qq(<input type="text"$p>);
}

# Pseudo-widget agglomerations

sub spreadsheet {
    my ($s, $args) = _arghash(@_);
    my $prefix = $args->{prefix};
    my $file = $args->{file};
    my $tab = SoMat::DB::Tab->new($file);
    return undef unless $tab->read;

    my $widths = [];
    my $row = 0;
    while (defined (my $r = $tab->get_row($row))) {
        for my $col (0 .. @$r - 1) {
            my $v = $r->[$col];
            my $len = length($v);
            $widths->[$col] = $len if $len > $widths->[$col];
        }
        ++$row;
    }
    $row = 0;
    my $t = $s->table;
    my $o = $t->begin('cellpadding'=>2);
    while (defined (my $r = $tab->get_row($row))) {
        $o .= $t->row;
        for my $col (0 .. @$r - 1) {
            $o .= $t->td($s->textfield('name'=>"$prefix$row\:$col",
                    'size'=>$widths->[$col] || 1, 'value'=>$r->[$col]));
        }
        ++$row;
    }
    $o .= $t->end;
    return $o;
}

sub end {
    return "</form>";
}

package SoMat::eDAQ::Web::Table;
@SoMat::eDAQ::Web::Table::ISA = qw(SoMat::eDAQ::Web);

my $DefaultRuleColor = '#808080';
my $DefaultRuleWidth = 1;

*_arghash = \&SoMat::eDAQ::Web::_arghash;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    $s->{attr} = (defined $_[0] && ref($_[0]) eq "HASH") ? $_[0] : { @_ };
    return $s;
}

sub begin {
    #my $s = shift;
    #my $attr = $s->{attr} = { @_ };
    my ($s, $args) = _arghash(@_);
    my $attr = $s->_arg_merge($s->{attr}, $args);

    $s->{row_bg_list} = $attr->{row_bg} if exists $attr->{row_bg};
    $s->{col_bg_list} = $attr->{col_bg} if exists $attr->{col_bg};
    $s->{row_bg_i} = 0;
    # KLUDGE fix border, cellpadding, cellspacing
    my $o = '';
    if (exists $attr->{rulewidth} || exists $attr->{rulecolor}) {
        my $width = $s->_argvalues($attr, 'width');
        my $rulebg = $s->_bgcolor((exists $attr->{rulecolor})
            ? $attr->{rulecolor} : $DefaultRuleColor);

        my $rulewidth = (exists $attr->{rulewidth})
            ? $attr->{rulewidth} : $DefaultRuleWidth;

        #$o .= qq(<table border=0 cellpadding=0 cellspacing=0$width$rulebg>\n);
        $o .= qq(<table$width>\n);
        $o .= qq(<tr><td>\n);

        my $table_attr = qq( width="100%" border="0" cellspacing="$rulewidth");
        $table_attr .= $s->_argvalues($attr, {'cellpadding'=>2},
            'cellpadding', 'bgcolor', 'class');
        $o .= qq(<table$table_attr>\n);
        $s->{end_table} = qq(</table></td></tr></table>\n);
    } else {
        my $defaults = {
            #'border' => 0,
            #'cellpadding' => 0,
            #'cellspacing' => 0
        };
        my $table_attr = $s->_argvalues($attr, $defaults, 'class',
            'border', 'cellpadding', 'cellspacing', 'width', 'bgcolor');
        $o .= qq(<table$table_attr>\n);
        $s->{end_table} = qq(</table>\n);
    }
    $s->bg();
    return $o;
}

sub end {
    my ($s) = @_;
    my $o = $s->end_row;
    return $o . $s->{end_table};
}

sub _bgcolor {
    my ($s, $color) = @_;
    return (defined $color) ? qq( bgcolor="$color") : "";
}

sub bg {
    my ($s, $color) = @_;
    $s->{cell_bg} = $s->_bgcolor($color);
}

sub _next_row_bg {
    my ($s) = @_;
    if (exists $s->{row_bg_list}) {
        my $bg;
        while (1) {
            $s->{row_bg_i} = 0 if $s->{row_bg_i} >= @{$s->{row_bg_list}};
            $bg = $s->{row_bg_list}->[$s->{row_bg_i}];
            last unless ref $bg;
            $s->{row_bg_list} = $bg;
            $s->{row_bg_i} = 0;
        }
        $s->row_bg($bg);
    }
    $s->{row_bg} = "";
    if ($s->{attr}{even_odd_shaded}) {
        $s->{row_bg} = ($s->{row_i} & 1) ? "" : qq{ class="even"};
    }
}

sub row_bg {
    my ($s, $color) = @_;
    $s->{row_bg} = $s->_bgcolor($color);
}

sub font {
    my ($s, $args) = _arghash(@_);
    #my $attr = $s->_argvalues($args, 'face', 'size', 'color');
    #$s->{cell_font} = ($attr) ? qq(<font$attr>) : "";
    #$s->{cell_unfont} = ($attr) ? qq(</font>) : "";
}

sub _alignment {
    #my ($s, $spec, $attr) = @_;
    my ($s, $spec, @attr) = @_;
    my $set = "";
    $set .= ' align="left"' if $spec =~ /l/i;
    $set .= ' align="center"' if $spec =~ /c/i;
    $set .= ' align="right"' if $spec =~ /r/i;
    $set .= ' valign="top"' if $spec =~ /t/i;
    $set .= ' valign="middle"' if $spec =~ /m/i;
    $set .= ' valign="bottom"' if $spec =~ /b/i;
    #$set .= " $attr" if defined $attr && $attr;
    if (@attr) {
        $set .= $s->_argvalues({ @attr },
            'class',
            'nowrap', 'rowspan', 'colspan', 'width', 'height', 'bgcolor');
    }
    return $set;
}

sub row {
    my ($s, $align_spec, @attr) = @_;
    my $attr = $s->_alignment($align_spec, @attr);
    my $o = $s->end_row;
    $s->_next_row_bg;
    $s->{in_row} = 1;
    return qq($o<tr$attr$s->{row_bg}>\n);
}

sub end_row {
    my ($s) = @_;
    if ($s->{in_row}) {
        ++$s->{row_i};
        ++$s->{row_bg_i};
        return qq(</tr>\n);
    }
    $s->{in_row} = 0;
    return '';
}

sub th {
    my ($s, $content, $align_spec, @attr) = @_;
    my $attr = $s->_alignment($align_spec, @attr);
    $content = '&nbsp;' unless defined $content;
    return qq(<th$attr$s->{cell_bg}>$content</th>\n);
}

sub td {
    my ($s, $content, $align_spec, @attr) = @_;
    my $attr = $s->_alignment($align_spec, @attr);
    $content = '&nbsp;' unless defined $content;
    return qq(<td$attr$s->{cell_bg}>$content</td>\n);
}

1;
