package SoMat::eDAQ::Web::Multipart;

my $MultipartNotice = <<'EOT';
* Multipart read support. Excerpted from CGI.pm 1.30 and heavily modified:
*
* Copyright 1995-1998 Lincoln D. Stein.  All rights reserved.
* It may be used and modified freely, but I do request that this copyright
* notice remain attached to the file.  You may modify this module as you 
* wish, but if you redistribute a modified version, please attach a note
* listing the modifications you have made.
*
* The most recent version of CGI.pm and complete docs are available at:
*   http://stein.cshl.org/WWW/software/CGI/
EOT

my $CRLF = "\015\012";

sub new {
    my ($package, $server, $boundary, $length) = @_;

    # Netscape seems to be a little bit unreliable
    # about providing boundary strings.
    my $boundary_read = 0;
    if ($boundary) {

        # Under the MIME spec, the boundary consists of the 
        # characters "--" PLUS the Boundary string

        # BUG: IE 3.01 on the Macintosh uses just the boundary -- not
        # the two extra hyphens. We do a special case here on the user-agent!
        $boundary = "--$boundary";
            # unless CGI::user_agent('MSIE\s+3\.0[12];\s*Mac|DreamPassport');

    } else { # otherwise we find it ourselves
        $boundary = $server->get_line;
        $length -= length($boundary);
        chomp($boundary);               # remove the CRLF
        $boundary_read++;
    }

    my $s = bless {
        server=>$server,
        length=>$length,
        boundary=>$boundary,
        buffer=>'',
        fillunit=>1024 * 4,
        total_bytes=>0,
    }, ref $package || $package;

    $s->{fillunit} = length($boundary)
        if length($boundary) > $s->{fillunit};

    # Read the preamble and the topmost (boundary) line plus the CRLF.
    unless ($boundary_read) {
        1 while $s->read(0);
    }
    die "Malformed multipart POST\n" if $s->eof;

    return $s;
}

sub readHeader {
    my ($s) = @_;
    my $end;
    my $ok = 0;

    do {
        $s->fillBuffer($s->{fillunit});
        $ok++ if ($end = index($s->{buffer}, "${CRLF}${CRLF}")) >= 0;
        $ok++ if $s->{buffer} eq '';
        return undef if !$ok && $s->{length} <= 0;
    } until $ok;

    my $header = substr($s->{buffer}, 0, $end + 2);
    substr($s->{buffer}, 0, $end + 4, '');

    # See RFC 2045 Appendix A and RFC 822 sections 3.4.8
    #   (Folding Long Header Fields), 3.4.3 (Comments)
    #   and 3.4.5 (Quoted-Strings).

    my $token = '[-\w!\#$%&\'*+.^_\`|{}~]';
    $header =~ s/$CRLF\s+/ /og;         # merge continuation lines

    my %return;
    while ($header =~ /($token+):\s+([^$CRLF]*)/mgox) {
        my ($field_name, $field_value) = ($1, $2);
        $return{lc($field_name)} = $field_value;
    }
    return \%return;
}

# This reads and returns the body as a single scalar value.
sub readBody {
    my ($s) = @_;
    my $data;
    my $body = '';
    while (defined($data = $s->read)) {
        $body .= $data;
    }
    return $body;
}

# This will read $bytes or until the boundary is hit, whichever happens
# first.  After the boundary is hit, we return undef.  The next read will
# skip over the boundary and begin reading again;
sub read {
    my ($s, $bytes) = @_;

    # default number of bytes to read
    $bytes = $bytes || $s->{fillunit};       
#print STDERR "Multipart read(), bytes=$bytes, length=$s->{length}\n";

    # Fill up our internal buffer in such a way that the boundary
    # is never split between reads.
    $s->fillBuffer($bytes);
#print STDERR "Multipart read(), fillBuffer done\n";

    # Find the boundary in the buffer (it may not be there).
    my $start = index($s->{buffer}, $s->{boundary});
    # protect against malformed multipart POST operations
    die "Malformed multipart POST\n" unless $start >= 0 || $s->{length} > 0;

#print STDERR "Multipart read(), start=$start\n";
    # If the boundary begins the data, then skip past it
    # and return undef.
    if ($start == 0) {

        # clear us out completely if we've hit the last boundary.
        if (index($s->{buffer}, "$s->{boundary}--") == 0) {
            $s->{buffer} = '';
            $s->{length} = 0;
            return undef;
        }

        # just remove the boundary.
        substr($s->{buffer}, 0, length($s->{boundary})) = '';
        $s->{buffer} =~ s/^\012\015?//;
        return undef;
    }

    my $bytesToReturn;
    if ($start > 0) {           # read up to the boundary
        $bytesToReturn = $start-2 > $bytes ? $bytes : $start;
    } else {    # read the requested number of bytes
        # leave enough bytes in the buffer to allow us to read
        # the boundary.  Thanks to Kevin Hendrick for finding
        # this one.
        $bytesToReturn = $bytes - (length($s->{boundary})+1);
    }
#print STDERR "Multipart read(), bytesToReturn=$bytesToReturn\n";

    my $returnval = substr($s->{buffer}, 0, $bytesToReturn);
    substr($s->{buffer}, 0, $bytesToReturn) = '';
    
#print STDERR "Multipart read(), returning\n";
    # If we hit the boundary, remove the CRLF from the end.
    return ($bytesToReturn == $start)
           ? substr($returnval, 0, -2) : $returnval;
}

# This fills up our internal buffer in such a way that the
# boundary is never split between reads
sub fillBuffer {
    my ($s, $bytes) = @_;
    return unless $s->{length};

    my $bufferLength = length($s->{buffer});
    my $toget = $bytes - $bufferLength + length($s->{boundary}) + 2;
    $toget = $s->{length} if $s->{length} < $toget;
#print STDERR "Multipart fillBuffer(), bytes=$bytes, total_bytes=$s->{total_bytes}, length=$s->{length}, toget=$toget\n";

    my $got = $s->{server}->rawread($s->{buffer}, $toget, $bufferLength);

    $s->{total_bytes} += $got;
    $s->{buffer} = '' unless defined $s->{buffer};

    die "Client aborted during multipart read.\n" unless $got;

    $s->{length} -= $got;
}

sub eof {
    my ($s) = @_;
    return 1 if length($s->{buffer}) == 0 && $s->{length} <= 0;
    return undef;
}

1;
