#!/usr/bin/perl
#
# eDAQ hardware spec file
#
# $Id: GIT $

use strict;

my $SECONDS = 0.5;

my $HLB_RATE = 2500;
my $HLB_CHANNELS = 16;

my $LLB_RATE = 10000;
my $LLB_CHANNELS = 8;

my $MPB_RATE = 2500;

my $CAT_RATE = 500;
my $CAT_CHANNELS = 128;

my $HLSS_RATE = 100000;
my $HLSS_CHANNELS = 2;

my $BRG_RATE = 100000;
my $BRG_CHANNELS = 2;

my $HLB_SIZE = $HLB_RATE * $HLB_CHANNELS * 2 * $SECONDS;
my $LLB_SIZE = $LLB_RATE * $LLB_CHANNELS * 2 * $SECONDS;
my $MPB_SIZE = $MPB_RATE * (6 + (4 * 8)) * $SECONDS;
my $VBUS_SIZE = 49152;
my $CAT_SIZE = $CAT_RATE * $CAT_CHANNELS * 2 * $SECONDS;
my $DIO_SIZE = 65536;
my $ITB_SIZE = 16384;
my $TCB_SIZE = 65536;
my $HLSS_SIZE = $HLSS_RATE * $HLSS_CHANNELS * 2 * $SECONDS;
my $BRG_SIZE = $BRG_RATE * $BRG_CHANNELS * 2 * $SECONDS;

sub device_fix {
    my ($node) = @_;
    if (-f '/eDAQ/.is-an-edaq-classic') {
        $node =~ s|^/dev/edaq/(\w+)/|/dev/$1|;
        $node =~ s|^/dev/mpbserial1$|/dev/mpbserial0|; # KLUDGE
    }
    return $node;
}

sub mpb_version {
    my $ver = `cat /etc/version/string`;
    chomp $ver;
    $ver =~ s/ build \d+$//;
    return $ver;
}

sub mpb_build {
    my $ver = `cat /etc/version/build`;
    chomp $ver;
    return $ver;
}

sub powermicro_version {
    my $version;
    for (my $i = 0; $i < 3; $i++) {
        $version = `pwrcmd 'VERSION?'`;
        chomp $version;
        if ($version =~ (/^0\|(.*)$/)) {
            $version = $1;
            $version =~ tr/A-Z/a-z/;
            last;
        } else {
            $version = "v0.1";
        }
        sleep 5;
    }
    return $version;
}

sub detect_vbms {
    require SoMat::eDAQ::Hardware::VBM;
    my ($db, $devref, $ret, $excitation, $exc_if_found, $descr,
        $prefix, $num_chs, $luns) = @_;
    foreach my $lun (@$luns) {
        $devref->send_packet("SET LUN=$lun " .
                             "INPUT_THRESHOLD_VOLTAGE=2500 " .
                             "SHUTDOWN=OFF");
        $devref->recv_cmd_response();
        for (my $channel = 0; $channel < $num_chs; ++$channel) {
            $devref->send_packet("SET LUN=$lun CHANNEL=$channel "
                . "VBM_CONTROL_PORT=0 VBM_ENABLE=NO $excitation");
            $devref->recv_cmd_response();
        }
    }
    select undef, undef, undef, 0.5; # KLUDGE tighten this up?
    my $port = 0x2000;
    foreach my $lun (@$luns) {
        for (my $channel = 0; $channel < $num_chs; ++$channel) {
            my $address = "LUN=$lun CHANNEL=$channel";
            my $present = 
                $devref->get_named_variable("VBM_PRESENT", 
                                            $address);
            if ($present eq 'YES') {
                if ($exc_if_found) {
                    $devref->send_packet("SET $address $exc_if_found");
                    $devref->recv_cmd_response();
                }
                $devref->send_packet("SET $address " .
                                     "VBM_CONTROL_PORT=$port " .
                                     "VBM_ENABLE=YES");
                $devref->recv_cmd_response();
                my $vbm = SoMat::eDAQ::Hardware::VBM->new($devref,
                                                          $port);
                my $serial = $vbm->run_command("SERNUM?");
                $serial ||= $ret->{serial};
                my $version = $vbm->run_command("VERSION?");
                $version =~ tr/A-Z/a-z/;
                my $build = $vbm->run_command("BUILDNUM?");
                my $appstate = $vbm->run_command("APPSTATE?");
                $db->add(description => $descr,
                         device => $ret->{device},
                         tcename => $prefix . $db->next_uid(),
                         serial => $serial,
                         version => $version,
                         build => $build,
                         interface => $appstate,
                         address => $ret->{address},
                         parent => $ret->{tcename},
                         module_address => $address);
                undef $vbm;
                $devref->send_packet("SET $address " .
                                     "VBM_CONTROL_PORT=0 " .
                                     "VBM_ENABLE=NO");
                $port++;
                $devref->recv_cmd_response();
            }
        }
    }
}

sub expmod_daughterboard {
    my ($ret, $devref) = @_;
    $ret->{db_serial} = $devref->get_named_variable("DB_SERIAL_NUMBER");
}

sub gps_command {
    my ($devref, $cmd) = @_;
    $devref->send_packet("SET LUN=GPS $cmd");
    $devref->recv_cmd_response;
}

{
    '^MSMPB.*' => [
        {
            description => 'Main Processor',
            tcename => 'MPB',
            device => device_fix('/dev/edaq/mpb/%d'),
            version => \&mpb_version,
            build => \&mpb_build,
            slot => 'hardware',
            server_buffer_hint => $MPB_SIZE,
        }, {
            description => 'MPB Serial Bus',
            tcename => 'MPBSer',
            device => device_fix('/dev/edaq/mpbserial/1'), # KLUDGEy
            slot => 'logical',
        }, {
            do_if => sub { -f "/eDAQ/.is-an-edaq-plus"; },
            description => 'Power Controller',
            tcename => 'Power',
            device => device_fix('/dev/nullish'),
            version => \&powermicro_version,
            slot => 'logical',
        },
    ],
    '^ELMPB.*' => [
        {
            description => 'Main Processor',
            tcename => 'MPB',
            device => device_fix('/dev/nullish'),
            version => \&mpb_version,
            build => \&mpb_build,
            slot => 'hardware',
        }, {
            description => 'MPB Serial Bus',
            tcename => 'MPBSer',
            device => device_fix('/dev/edaq/mpbserial/1'), # KLUDGEy
            slot => 'logical',
        },
    ],
    '^ELPWR.*' => [
        {
            description => 'Power Controller',
            tcename => 'Power',
            device => device_fix('/dev/nullish'),
            version => \&powermicro_version,
            slot => 'hardware',
        },
    ],
    '^MSHLB.*' => [
        {
            description => 'High Level',
            tcename => 'HiLev_%d',
            device => device_fix('/dev/edaq/hlb/%d'),
            slot => 'hardware',
            server_buffer_hint => $HLB_SIZE,
        }, {
            description => 'Vehicle Bus',
            tcename => 'VBus_%d',
            device => device_fix('/dev/edaq/vbus/%d'),
            version => sub { 
                my ($device) = @_;
                my $verstr;
                sysopen(VBUS, $device, 2);
                ioctl(VBUS, 1, my $junk);
                my $count = sysread(VBUS, $verstr, 5);
                close VBUS;
                my $version = ord substr($verstr, 4, 1);
                return sprintf("v%d.%d", 
                    ($version & 0xf0) >> 4, $version & 0x0f);
            },
            slot => 'logical',
            server_buffer_hint => $VBUS_SIZE,
        },
    ],
    '^MSHLA.*' => [
        {
            description => 'High Level',
            tcename => 'HiLev_%d',
            device => device_fix('/dev/edaq/hlb/%d'),
            slot => 'hardware',
            server_buffer_hint => $HLB_SIZE,
        }
    ],
    '^MSLLB.*' => [
        {
            description => 'Low Level',
            tcename => 'LoLev_%d',
            device => device_fix('/dev/edaq/llb/%d'),
            slot => 'hardware',
            server_buffer_hint => $LLB_SIZE,
        },
    ],
    '^(MS|EL)CAT.*' => [
        {
            description => 'Caterpillar ECM',
            tcename => 'CatE_%d',
            slot => 'hardware',
            kernel_buffer_hint => $CAT_SIZE,
            server_buffer_hint => $CAT_SIZE,
        },
    ],
    '^MSITB.*' => [
        {
            description => 'Isolated Thermocouple',
            tcename => 'IsoTC_%d',
            slot => 'hardware',
            extra => sub {
                my ($ret, $devref) = @_;
                my $type = $devref->get_named_variable("THERMOCOUPLE_TYPE", 
                    "LUN=ISO_8_TC CHANNEL=0");
                $ret->{thermocouple_type} = $type;
            },
            kernel_buffer_hint => $ITB_SIZE,
            server_buffer_hint => $ITB_SIZE,
        },
    ],
    '^(MS|EL)TCB.*' => [
        {
            description => 'Thermocouple',
            tcename => 'StdTC_%d',
            slot => 'hardware',
            kernel_buffer_hint => $TCB_SIZE,
            server_buffer_hint => $TCB_SIZE,
            extra => sub {
                my ($ret, $devref) = @_;
                if ($ret->{serial} =~ /^EL/) {
                    $devref->send_packet("SET PLL_FREQUENCY=80");
                    $devref->recv_cmd_response;
                }
            },
        },
    ],
    '^(MS|EL)HLS.*' => [
        {
            description => 'High Level SS',
            tcename => 'HLSS_%d',
            slot => 'hardware',
            kernel_buffer_hint => $HLSS_SIZE,
            server_buffer_hint => $HLSS_SIZE,
        },
    ],
    '^(MS|EL)BRG.*' => [
        {
            description => 'Bridge',
            tcename => 'Brg_%d',
            slot => 'hardware',
            kernel_buffer_hint => $BRG_SIZE,
            server_buffer_hint => $BRG_SIZE,
            extra => \&expmod_daughterboard,
        },
    ],
    '^MSMDI.*' => [
        {
            description => 'MegaDAC Interface',
            tcename => 'MDI_%d',
            slot => 'hardware',
        },
    ],
    '^MSUDB.*' => [
        {
            description => 'Universal Digital',
            tcename => 'UDB_%d',
            slot => 'hardware',
            kernel_buffer_hint => $DIO_SIZE,
            server_buffer_hint => $DIO_SIZE,
            extra => \&expmod_daughterboard,

        },
    ],
    '^(MS|EL)DIO.*' => [
        {
            description => 'Digital I/O',
            tcename => 'DIO_%d',
            slot => 'hardware',
            kernel_buffer_hint => $DIO_SIZE,
            server_buffer_hint => $DIO_SIZE,
            after => sub {
                my ($db, $devref, $ret, $name, $num_chs, @luns) = @_;
                detect_vbms($db, $devref, $ret,
                            "EXCITATION=5", "EXCITATION=12",
                            "Vehicle Bus Module", "VBM_", 2,
                            [ "IO_BANK_A", "IO_BANK_B", "IO_BANK_C" ]);
            },
        },
        {
            do_if => sub {
                my ($ret, $devref) = @_;
                my $model = $devref->get_named_variable("MODEL", "LUN=GPS");
                return $model; # don't do if GPS not installed
            },
            description => 'GPS',
            tcename => 'GPS_%d',
            slot => 'logical',
            extra => sub {
                my ($ret, $devref) = @_;
                my $model = $devref->get_named_variable("MODEL", "LUN=GPS");
                $ret->{gps_model} = $model;
            },
        },
    ],
    '^(MS|EL)COM.*' => [
        {
            description => 'COM',
            tcename => 'COM_%d',
            slot => 'hardware',
            kernel_buffer_hint => $DIO_SIZE,
            server_buffer_hint => $DIO_SIZE,
            after => sub {
                my ($db, $devref, $ret, $name, $num_chs, @luns) = @_;
                detect_vbms($db, $devref, $ret, "EXCITATION_VOLTAGE=0", '',
                            "CAN Vehicle Bus", "CAN_", 3, [ "CAN" ]);
                detect_vbms($db, $devref, $ret, "EXCITATION_VOLTAGE=12", '',
                            "Vehicle Bus Module", "VBM_", 1, [ "VBM" ]);
            },
        },
        {
            do_if => sub {
                my ($ret, $devref) = @_;
                my $model = $devref->get_named_variable("MODEL", "LUN=GPS");
                return $model; # don't do if GPS not installed
            },
            description => 'GPS',
            tcename => 'GPS_%d',
            slot => 'logical',
            extra => sub {
                my ($ret, $devref) = @_;
                #open(my $DBG, '>', '/tmp/hwquery.dbg');
                my $gps_port = 0x200;

                my $model = $devref->get_named_variable("MODEL", "LUN=GPS");

                my $orig = {
                    EXCITATION_VOLTAGE=>5, # Detection default for everything
                    BAUD_RATE=>38400,      # Default for Garmin, EGPS-200
                    MODEL=>$model,
                };

                if ($orig->{MODEL} =~ /^leadtek/i) {
                    $orig->{BAUD_RATE} = 4800;
                }

                my $new = { %$orig };

                my $setup = [
                    "EXCITATION_VOLTAGE=$orig->{EXCITATION_VOLTAGE}",
                    "BAUD_RATE=$orig->{BAUD_RATE}",
                    "LATENCY_PERIOD=20",
                    "TERMINATOR=10", # LF
                    sprintf("RX_PORT=%d", $gps_port),
                    "TIMESTAMP_COUNTER=0",
                    "RUN=ON",
                ];

                for my $cmd (@$setup) {
                    #print $DBG "SETUP: $cmd\n";
                    gps_command($devref, $cmd);
                }

                my $gps_stream = '';

                ITER: for my $i (0..50) {
                    #print $DBG "ITER $i\n";
                    my $pkt = $devref->recv_packet($gps_port);
                    last ITER unless defined $pkt;
                    my ($ts, $pkt_type, $payload) = unpack('Vva*', $pkt);
                    next if $pkt_type != 1;
                    $gps_stream .= $payload;
                    my $hex = unpack('H*', $payload);
                    $hex =~ s/(..)/$1 /g;
                    $payload =~ s/(.)/ $1 /g;
                    #print $DBG "GPS_PKT (ts=$ts):\n    $hex\n    $payload\n";
                    while ($gps_stream =~ s/(.*?)\x0D\x0A//) {
                        my $msg = $1;
                        #next if $msg =~ /[\x00-\x1F\x80-\xFF]/;
                        if ($msg =~ /^\$(PSMTID\,(.*?))\*(..)$/) {
                            my ($body, $id, $checksum) = ($1, $2, $3);
                            #print $DBG "body=$body, id=$id, cs=$checksum\n";
                            my $sum = 0;
                            map { $sum ^= ord($_) } split(//, $body);
                            if ($sum == hex($checksum)) {
                                $new->{MODEL} = $id;
                                #print $DBG "GOT GPS MODEL $id\n";
                                last ITER;
                            } else {
                                #printf $DBG "BAD CHECKSUM %.02X\n", $sum;
                            }
                        }
                    }
                }

                gps_command($devref, "RUN=OFF");

                if ($new->{MODEL} eq 'EGPS-200') {
                    $new->{BAUD_RATE} = 460800;
                    $new->{EXCITATION_VOLTAGE} = 12;
                }

                for my $id (qw(EXCITATION_VOLTAGE BAUD_RATE)) {
                    if ($orig->{$id} ne $new->{$id}) {
                        #print $DBG "RE-SETUP: $id=$new->{$id}\n";
                        gps_command($devref, "$id=$new->{$id}")
                    }
                }
                #print $DBG "RE-SETUP DONE\n";

                $ret->{gps_model} = $new->{MODEL};
                $ret->{gps_baud_rate} = $new->{BAUD_RATE};
            },
        },
    ],
    '__UNKNOWN__' => [
        {
            description => 'Unknown',
            tcename => 'Unknown_%d',
        },
    ],
};
